/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "compiler.h"

#include "xf86Pci.h"
#include "xf86PciInfo.h"

#include "vgaHW.h"

#include "savage_driver.h"

static void
SavageI2CPutBits(I2CBusPtr b, int clock,  int data)
{
    ScrnInfoPtr pScrn = (ScrnInfoPtr)(xf86Screens[b->scrnIndex]);
    SavagePtr psav = SAVPTR(pScrn);
    vgaHWPtr hwp;
    int vgaIOBase;
    unsigned char reg = 0x10;

    hwp = VGAHWPTR(pScrn);
    vgaHWGetIOBase(hwp);
    vgaIOBase = hwp->IOBase;

    if(clock) reg |= 0x1;
    if(data)  reg |= 0x2;

    OutI2CREG(reg,psav->DDCPort);
}

static void
SavageI2CGetBits(I2CBusPtr b, int *clock, int *data)
{
    ScrnInfoPtr pScrn = (ScrnInfoPtr)(xf86Screens[b->scrnIndex]);
    SavagePtr psav = SAVPTR(pScrn);
    vgaHWPtr hwp;
    int vgaIOBase;
    unsigned char reg;

    hwp = VGAHWPTR(pScrn);
    vgaHWGetIOBase(hwp);
    vgaIOBase = hwp->IOBase;

    InI2CREG(reg,psav->DDCPort);

    *clock = reg & 0x4;
    *data = reg & 0x8;
}

Bool
SavageI2CInit(ScrnInfoPtr pScrn)
{
    SavagePtr psav = SAVPTR(pScrn);
    I2CBusPtr I2CPtr;

    I2CPtr = xf86CreateI2CBusRec();
    if(!I2CPtr) return FALSE;

    psav->I2C = I2CPtr;

    I2CPtr->BusName    = "I2C bus";
    I2CPtr->scrnIndex  = pScrn->scrnIndex;
    I2CPtr->I2CPutBits = SavageI2CPutBits;
    I2CPtr->I2CGetBits = SavageI2CGetBits;

    if (!xf86I2CBusInit(I2CPtr))
	return FALSE;

    return TRUE;
}

static void
SavageDVIPutBits(I2CBusPtr b, int clock,  int data)
{
    ScrnInfoPtr pScrn = (ScrnInfoPtr)(xf86Screens[b->scrnIndex]);
    SavagePtr psav = SAVPTR(pScrn);
    vgaHWPtr hwp;
    int vgaIOBase;
    unsigned char reg = 0x10;

    hwp = VGAHWPTR(pScrn);
    vgaHWGetIOBase(hwp);
    vgaIOBase = hwp->IOBase;
    if(clock) reg |= 0x1;
    if(data)  reg |= 0x2;

    OutI2CREG(reg,psav->I2CPort);
}

static void
SavageDVIGetBits(I2CBusPtr b, int *clock, int *data)
{
    ScrnInfoPtr pScrn = (ScrnInfoPtr)(xf86Screens[b->scrnIndex]);
    SavagePtr psav = SAVPTR(pScrn);
    vgaHWPtr hwp;
    int vgaIOBase;
    unsigned char reg;

    hwp = VGAHWPTR(pScrn);
    vgaHWGetIOBase(hwp);
    vgaIOBase = hwp->IOBase;
    InI2CREG(reg,psav->I2CPort);

    *clock = reg & 0x4;
    *data = reg & 0x8;
}

Bool
SavageDVIInit(ScrnInfoPtr pScrn)
{
    SavagePtr psav = SAVPTR(pScrn);
    I2CBusPtr DVIPtr;

    DVIPtr = xf86CreateI2CBusRec();
    if(!DVIPtr)
        return FALSE;

    psav->DVI = DVIPtr;

    DVIPtr->BusName    = "DVI bus";
    DVIPtr->scrnIndex  = pScrn->scrnIndex;
    DVIPtr->I2CPutBits = SavageDVIPutBits;
    DVIPtr->I2CGetBits = SavageDVIGetBits;

    if (!xf86I2CBusInit(DVIPtr))
        return FALSE;

    return TRUE;
}


