/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#ifndef SAVAGEVB_INC
#define SAVAGEVB_INC

#include "vb.h"
#include "types.h"

/* 
 *  color type for the vertex data
 *  we probably want to use an internal datatype here?
 */
typedef struct {
  GLubyte	blue;
  GLubyte       green;
  GLubyte       red;
  GLubyte       alpha;
} savage_color;


/* A basic fixed format vertex to kick things off.  Move to dynamic
 * layouts later on.  (see also the savage_full_vertex struct in
 * savage_3d_reg.h)
 */
typedef struct {
   GLfloat x,y,z;	        /* coordinates in screen space*/
   GLfloat oow;			/* reciprocal homogeneous w */
   savage_color color;		/* vertex color */
   savage_color specular;		/* specular color, alpha is fog */
   GLfloat tu0,tv0;		/* texture 0 */
   GLfloat tu1,tv1;		/* texture 1 */
} savage_vertex;


/* Unfortunately only have assembly for 16-stride vertices.
 */
union savage_vertex_t {
   savage_vertex v;
   float f[16];
   GLuint ui[16];
};

typedef union savage_vertex_t savageVertex;
typedef union savage_vertex_t *savageVertexPtr;

struct savage_vertex_buffer_t {
   GLvector1ui clipped_elements;
   savageVertexPtr verts;
   int last_vert;
   GLuint *primitive;
   GLuint *next_primitive;
   void *vert_store;
   GLuint size;
};
   
typedef struct savage_vertex_buffer_t *savageVertexBufferPtr;


#define SAVAGE_CONTEXT(ctx)    ((savageContextPtr)(ctx->DriverCtx))
#define SAVAGE_DRIVER_DATA(vb) ((savageVertexBufferPtr)((vb)->driver_data))


#define SAVAGE_SPEC_BIT       0x1
#define SAVAGE_FOG_BIT	    0x2
#define SAVAGE_ALPHA_BIT      0x4	/* GL_BLEND, not used */
#define SAVAGE_TEX1_BIT       0x8
#define SAVAGE_TEX0_BIT       0x10	
#define SAVAGE_RGBA_BIT       0x20
#define SAVAGE_WIN_BIT        0x40


extern void savageChooseRasterSetupFunc(GLcontext *ctx);
extern void savagePrintSetupFlags(char *msg, GLuint flags );
extern void savageDDDoRasterSetup( struct vertex_buffer *VB );
extern void savageDDPartialRasterSetup( struct vertex_buffer *VB );
extern void savageDDCheckPartialRasterSetup( GLcontext *ctx, 
					   struct gl_pipeline_stage *d );

extern void savageDDViewport( GLcontext *ctx, 
			   GLint x, GLint y, 
			   GLsizei width, GLsizei height );

extern void savageDDDepthRange( GLcontext *ctx, 
			     GLclampd nearval, GLclampd farval );


extern void savageDDUnregisterVB( struct vertex_buffer *VB );
extern void savageDDRegisterVB( struct vertex_buffer *VB );
extern void savageDDResizeVB( struct vertex_buffer *VB, GLuint size );

extern void savageDDSetupInit( void );

#if 1
extern struct timeval tv_s,tv_f;
#endif

#endif
