/*
 * Copyright 2003 VIA Technologies, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*************************************************************************
** File libSavageXvMC.c
**
** Authors:
**	John Zhao <jzhao@s3graphics.com>
**
**
***************************************************************************/
/* $XFree86: xc/lib/XvMC/hw/savage/savageXvMC.c,v 1.7 2001/12/18 04:23:55 dawes Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>

#include <sys/ioctl.h>
#include <Xlibint.h>
#include <fourcc.h>
#include <Xv.h>
#include <Xvlib.h>
#include <XvMC.h>
#include <XvMClib.h>
#include "s3data.h"
#include "savageXvMC.h"
#include "savagehwmc.h"

static int error_base;
static int event_base;
static int framenum=0;
#define		GET_FSTATUS(pSavageXvMC)    0
#define		GET_RSTATUS(pSavageXvMC)    0
/***************************************************************************
// Function: free_privContext
// Description: Free's the private context structure if the reference
//  count is 0.
***************************************************************************/
void savage_free_privContext(savageXvMCContext *pSavageXvMC) {

/*  SAVAGE_LOCK(pSavageXvMC,DRM_LOCK_QUIESCENT);*/


  pSavageXvMC->ref--;
  if(!pSavageXvMC->ref) {
    drmUnmap(pSavageXvMC->Surfaces.address,pSavageXvMC->Surfaces.size);
    drmUnmap(pSavageXvMC->DCTBlocks.address,pSavageXvMC->DCTBlocks.size);
    drmUnmap(pSavageXvMC->Aperture.address,pSavageXvMC->Aperture.size);
    drmUnmap(pSavageXvMC->MMIO.address,pSavageXvMC->MMIO.size);
    drmClose(pSavageXvMC->fd);

    free(pSavageXvMC);
  }

/*  SAVAGE_UNLOCK(pSavageXvMC);*/
}

/***************************************************************************
// Function: XvMCCreateContext
// Description: Create a XvMC context for the given surface parameters.
// Arguments:
//   display - Connection to the X server.
//   port - XvPortID to use as avertised by the X connection.
//   surface_type_id - Unique identifier for the Surface type.
//   width - Width of the surfaces.
//   height - Height of the surfaces.
//   flags - one or more of the following
//	XVMC_DIRECT - A direct rendered context is requested.
//
// Notes: surface_type_id and width/height parameters must match those
//	  returned by XvMCListSurfaceTypes.
// Returns: Status
***************************************************************************/
Status XvMCCreateContext(Display *display, XvPortID port,
			 int surface_type_id, int width, int height, int flags,
			 XvMCContext *context) {
  savageXvMCContext *pSavageXvMC;
  char busIdString[10];
  int priv_count;
  uint *priv_data;
  uint magic;
  Status ret;
  int major, minor;
  int i;
#ifdef XVMCLOG
  printf("Start XvMCCreateContext\n");
#endif
  /* Verify Obvious things first */
  if(context == NULL) {
    return XvMCBadContext;
  }

  if(!flags & XVMC_DIRECT) {
    /* Indirect */
    printf("Indirect Rendering not supported!\nUsing Direct.");
  }

#if 0
  /* Limit use to root for now */
  if(geteuid()) {
    printf("Use of XvMC on savage is currently limited to root\n");
    return BadAccess;
  }
#endif

  /* FIXME: Check $DISPLAY for legal values here */

  context->surface_type_id = surface_type_id;
  context->width = (unsigned short)width;
  context->height = (unsigned short)height;
  context->flags = flags;
  context->port = port;
  /*
     Width, Height, and flags are checked against surface_type_id
     and port for validity inside the X server, no need to check
     here.
  */

  /* Allocate private Context data */
  context->privData = (void *)malloc(sizeof(savageXvMCContext));
  if(!context->privData) {
    printf("Unable to allocate resources for XvMC context.\n");
    return BadAlloc;
  }
  pSavageXvMC = (savageXvMCContext *)context->privData;
  memset((char *)pSavageXvMC, 0, sizeof(savageXvMCContext));

  /* Verify the XvMC extension exists */
  if(! XvMCQueryExtension(display, &event_base,
			  &error_base)) {
    printf("XvMC Extension is not available!\n");
    return BadAlloc;
  }
  /* Verify XvMC version */
  ret = XvMCQueryVersion(display, &major, &minor);
  if(ret) {
    printf("XvMCQuery Version Failed, unable to determine protocol version\n");
  }
  /* FIXME: Check Major and Minor here */

  /* Check for drm */
  if(! drmAvailable()) {
    printf("Direct Rendering is not avilable on this system!\n");
    return BadAlloc;
  }

  /*
     Build the Attribute Atoms, and Initialize the ones that exist
     in Xv.
  */
  pSavageXvMC->xv_colorkey = XInternAtom(display,"XV_COLORKEY",0);
  if(!pSavageXvMC->xv_colorkey) {
    return XvBadPort;
  }
  ret = XvGetPortAttribute(display,port,pSavageXvMC->xv_colorkey,
			   &pSavageXvMC->colorkey);
  if(ret) {
    return ret;
  }
  pSavageXvMC->xv_brightness = XInternAtom(display,"XV_BRIGHTNESS",0);
  pSavageXvMC->xv_saturation = XInternAtom(display,"XV_SATURATION",0);
  pSavageXvMC->xv_contrast = XInternAtom(display,"XV_CONTRAST",0);
  pSavageXvMC->xv_hue = XInternAtom(display,"XV_HUE",0);
  pSavageXvMC->brightness = 0;
  pSavageXvMC->saturation = 128;
  pSavageXvMC->contrast = 128;
  pSavageXvMC->hue = 0;
  pSavageXvMC->port = port;
  pSavageXvMC->image = NULL;
  pSavageXvMC->gc = 0;
  pSavageXvMC->data = 0;


  /* Open DRI Device */
  if((pSavageXvMC->fd = drmOpen("savage",NULL)) < 0) {
    printf("DRM Device for savage could not be opened.\n");
    free(busIdString);
    free(pSavageXvMC);
    return BadAccess;
  } /* !pSavageXvMC->fd */

  /* Get magic number and put it in privData for passing */
  drmGetMagic(pSavageXvMC->fd,&magic);
  context->flags = (uint)magic;

  /*
    Pass control to the X server to create a drmContext for us and
    validate the with/height and flags.
  */
  if((ret = _xvmc_create_context(display, context, &priv_count, &priv_data))) {
    printf("Unable to create XvMC Context.\n");
    return ret;
  }

  /*
     X server returns a structure like this:
     drmContext
     fbBase
     OverlayOffset
     OverlaySize
     SurfacesOffset
     SurfacesSize
     busIdString = 9 char + 1
  */

  if(priv_count != 20) {
    printf("_xvmc_create_context() returned incorrect data size!\n");
    printf("\tExpected 9, got %d\n",priv_count);
    _xvmc_destroy_context(display, context);
    free(pSavageXvMC);
    return BadAlloc;
  }

  pSavageXvMC->drmcontext = priv_data[0];
  pSavageXvMC->fb_base = priv_data[1];

  pSavageXvMC->MMIO.handle = priv_data[2];
  pSavageXvMC->MMIO.size = priv_data[3];

  pSavageXvMC->Surfaces.handle = priv_data[4];
  pSavageXvMC->Surfaces.offset = priv_data[5];
  pSavageXvMC->Surfaces.size = priv_data[6];

  pSavageXvMC->DCTBlocks.handle = priv_data[7];
  pSavageXvMC->DCTBlocks.offset = priv_data[8];
  pSavageXvMC->DCTBlocks.size = priv_data[9];

  pSavageXvMC->Aperture.handle = priv_data[10];
  pSavageXvMC->Aperture.size = priv_data[11];

  pSavageXvMC->depth = priv_data[12];
  pSavageXvMC->frameX0 = priv_data[13];
  pSavageXvMC->frameY0 = priv_data[14];
  pSavageXvMC->IOBase = priv_data[15];
  pSavageXvMC->displayWidth = priv_data[16];

#ifdef XVMCLOG
  printf("depth = %d, widht=%d\n", pSavageXvMC->depth, pSavageXvMC->displayWidth);
#endif

  strncpy(pSavageXvMC->busIdString,(char *)&priv_data[17],9);
  pSavageXvMC->busIdString[9] = '\0';

  /* Must free the private data we were passed from X */
  free(priv_data);

  /* Initialize private context values */
  pSavageXvMC->current = 0;
  pSavageXvMC->lock = 0;
  pSavageXvMC->last_flip = 0;
  pSavageXvMC->dual_prime = 0;

#ifdef XVMCLOG
  /* Map MMIO Register */
  printf("fd =%x mmio_handle=%x mmio_size = %x\n",  pSavageXvMC->fd, pSavageXvMC->MMIO.handle,
						   pSavageXvMC->MMIO.size);

  printf("fd =%x surface_handle=%x surface_size = %x\n",  pSavageXvMC->fd, pSavageXvMC->Surfaces.handle,
						   pSavageXvMC->Surfaces.size);

#endif

  printf("fd =%x block_handle=%x block_size = %x\n",  pSavageXvMC->fd, pSavageXvMC->DCTBlocks.handle,
						   pSavageXvMC->DCTBlocks.size);

  printf("fd =%x mmio_handle=%x mmio_size = %x\n",  pSavageXvMC->fd, pSavageXvMC->Aperture.handle,
						   pSavageXvMC->Aperture.size);
  if(drmMap(pSavageXvMC->fd,pSavageXvMC->MMIO.handle,
	    pSavageXvMC->MMIO.size,&(pSavageXvMC->MMIO.address)) < 0) {
    printf("Unable to map MMIO Register.\n");
    _xvmc_destroy_context(display, context);
    free(pSavageXvMC);
    return BadAlloc;
  } /* drmMap() < 0 */

  pSavageXvMC->BCIBaseAddr = (unsigned int)((unsigned int)pSavageXvMC->MMIO.address + 0x00010000);

  /* Map Surfaces */
  if(drmMap(pSavageXvMC->fd,pSavageXvMC->Surfaces.handle,
	    pSavageXvMC->Surfaces.size,&(pSavageXvMC->Surfaces.address)) < 0) {
    printf("Unable to map XvMC Surfaces.\n");
    _xvmc_destroy_context(display, context);
    free(pSavageXvMC);
    return BadAlloc;
  } /* drmMap() < 0 */

  /* Map DCT Block */
  if(drmMap(pSavageXvMC->fd,pSavageXvMC->DCTBlocks.handle,
	    pSavageXvMC->DCTBlocks.size,&(pSavageXvMC->DCTBlocks.address)) < 0) {
    printf("Unable to map XvMC DCTBlock.\n");
    _xvmc_destroy_context(display, context);
    free(pSavageXvMC);
    return BadAlloc;
  }

  /* Map Aperture Block */
  if(drmMap(pSavageXvMC->fd,pSavageXvMC->Aperture.handle,
	    pSavageXvMC->Aperture.size,&(pSavageXvMC->Aperture.address)) < 0) {
    printf("Unable to map XvMC Aperture.\n");
    _xvmc_destroy_context(display, context);
    free(pSavageXvMC);
    return BadAlloc;
  }

  pSavageXvMC->lpSymbolBuffer = (unsigned int *)malloc(0x100000);

  if(pSavageXvMC->lpSymbolBuffer == NULL) {
    printf("Unable to allocate symbol buffer.\n");
    _xvmc_destroy_context(display, context);
    free(pSavageXvMC);
    return BadAlloc;
  }

  /*
    There is a tiny chance that someone was using the overlay and
    issued a flip that hasn't finished. To be 100% sure I'll just
    take the lock and sleep for the worst case time for a flip.
  */

  pSavageXvMC->ref = 1;


  /* AGP memeory */
  for(i = 0; i < 4; i++)
  {
      pSavageXvMC->DCTBlock[i].offset = (unsigned int)((unsigned int)pSavageXvMC->DCTBlocks.address +
					pSavageXvMC->DCTBlocks.offset + i * 1024 * 1024);
      pSavageXvMC->DCTBlock[i].used = 0;
  }

  pSavageXvMC->FlipBuff[0] = (unsigned int)pSavageXvMC->Surfaces.offset;

  for(i = 1; i < MAXFLIPBUFF; i++)
  {
      /* Physical address */
      pSavageXvMC->FlipBuff[i] = (unsigned int)(pSavageXvMC->FlipBuff[i - 1] + 786*576*2 + 2048 + 2047) & 0xfffff800;
  }

  /* Physical address */
  pSavageXvMC->dwSubPicMixer = (unsigned int)(pSavageXvMC->FlipBuff[MAXFLIPBUFF - 1] + 786*576*2 + 2048 + 2047) & 0xfffff800;

  /* Physical Address */
  pSavageXvMC->SubPicBuffOffset = (unsigned int)((pSavageXvMC->dwSubPicMixer + 786*576*3 + 2048 + 2047) & 0xfffff800);

  /* Linear Address*/
  pSavageXvMC->SubPicBuff = (unsigned int)(pSavageXvMC->SubPicBuffOffset - pSavageXvMC->Surfaces.offset +							       (unsigned int)pSavageXvMC->Surfaces.address);


  pSavageXvMC->SubPicPaletteOffset = (unsigned int)(pSavageXvMC->SubPicBuffOffset + 786*576 + 2048 + 2047) & 0xfffff800;

  pSavageXvMC->SubPicPalette = (unsigned int)(pSavageXvMC->SubPicPaletteOffset - pSavageXvMC->Surfaces.offset
						+ (unsigned int)pSavageXvMC->Surfaces.address);


  pSavageXvMC->dwMixerStride = 3954;
  pSavageXvMC->dwTiledSurfReg = TILED_SURFACE_REGISTER_2;
  pSavageXvMC->fpTargetAperture =  (unsigned int)((unsigned int)pSavageXvMC->Aperture.address + 0x1000000 * 2);
  pSavageXvMC->usEventTag = 0;

#ifdef XVMCLOG
  printf("Surface offset=%x\n",pSavageXvMC->Surfaces.offset);
  printf("FlipBuff[0]=%x\n", pSavageXvMC->FlipBuff[0]);
  printf("FlipBuff[1]=%x\n", pSavageXvMC->FlipBuff[1]);
  printf("dwSubPicMixer=%x\n", pSavageXvMC->dwSubPicMixer);
  printf("SubPicBuffOffset=%x\n", pSavageXvMC->SubPicBuffOffset);
  printf("PaletteOffset=%x\n", pSavageXvMC->SubPicPaletteOffset);
#endif
#ifdef XVMCLOG
  printf("End XvMCCreateContext\n");
#endif
  return Success;

}

/***************************************************************************
// Function: XvMCDestroyContext
// Description: Destorys the specified context.
//
// Arguments:
//   display - Specifies the connection to the server.
//   context - The context to be destroyed.
//
// Returns: Status
***************************************************************************/
Status XvMCDestroyContext(Display *display, XvMCContext *context) {
  savageXvMCContext *pSavageXvMC;
#ifdef XVMCLOG
  printf("Start XvMCDestroyContext\n");
#endif
  if(context == NULL) {
    return (error_base + XvMCBadContext);
  }
  if(context->privData == NULL) {
    return (error_base + XvMCBadContext);
  }
  pSavageXvMC = (savageXvMCContext *)context->privData;

  /* Turn off the overlay */
  SAVAGE_LOCK(pSavageXvMC,0);
  MCS3ResetStreams(pSavageXvMC);
  SAVAGE_UNLOCK(pSavageXvMC);

  /* Pass Control to the X server to destroy the drmContext */
  _xvmc_destroy_context(display, context);

  free(pSavageXvMC->lpSymbolBuffer);
  free(pSavageXvMC->data);
  XFree(pSavageXvMC->image);
  savage_free_privContext(pSavageXvMC);
  context->privData = NULL;

#ifdef XVMCLOG
  printf("End XvMCDestroyContext\n");
#endif
  return Success;
}


/***************************************************************************
// Function: XvMCCreateSurface
***************************************************************************/
Status XvMCCreateSurface( Display *display, XvMCContext *context,
			  XvMCSurface *surface) {
  savageXvMCContext *pSavageXvMC;
  savageXvMCSurface *pSavageSurface;
  int priv_count;
  uint *priv_data;
  Status ret;
#ifdef XVMCLOG
  printf("Start XvMCCreateSurface\n");
#endif
  if((surface == NULL) || (context == NULL) || (display == NULL)){
    return BadValue;
  }

  pSavageXvMC = (savageXvMCContext *)context->privData;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadContext);
  }


  surface->privData = (savageXvMCSurface *)malloc(sizeof(savageXvMCSurface));
  if(!surface->privData) {
    return BadAlloc;
  }
  pSavageSurface = (savageXvMCSurface *)surface->privData;

  /* Initialize private values */
  pSavageSurface->privContext = pSavageXvMC;
  pSavageSurface->last_render = 0;
  pSavageSurface->last_flip = 0;
  pSavageSurface->second_field = 0;

  if((ret = _xvmc_create_surface(display, context, surface,
				&priv_count, &priv_data))) {
    free(pSavageSurface);
    printf("Unable to create XvMCSurface.\n");
    return ret;
  }

  /*
    _xvmc_create_subpicture returns 2 uints with the offset into
    the DRM map for the Y surface and UV surface.
  */
  if(priv_count != 1) {
    printf("_xvmc_create_surface() return incorrect data size.\n");
    printf("Expected 2 got %d\n",priv_count);
    free(priv_data);
    free(pSavageSurface);
    return BadAlloc;
  }
  /* Data == Client Address, offset == Physical address offset */
  pSavageSurface->data = pSavageXvMC->Surfaces.address;
  pSavageSurface->handle = pSavageXvMC->Surfaces.handle;
  pSavageSurface->offsets[0] = priv_data[0];

  pSavageSurface->pitch = surface->width * 3 / 2;

  /* Free data returned from xvmc_create_surface */
  free(priv_data);

  pSavageSurface->wWidth = (unsigned short)surface->width;
/*  surface->height=480;*/
  pSavageSurface->wHeight = (unsigned short)surface->height;

  pSavageSurface->dwLinearPitch = (surface->width) * 3 / 2;
  pSavageXvMC->dwDummyTile = ((surface->width + 63) & 0xffffffc0) - surface->width;
  pSavageXvMC->dwStride = ((((surface->width + 63) & 0xffffffc0) >> 6) << 8) - 135;

  if(pSavageXvMC->dwDummyTile)
      pSavageXvMC->dwFlags |= MB_DUMMY_BLOCK;

  pSavageXvMC->dwDummyTile = pSavageXvMC->dwDummyTile << 1;

#ifdef XVMCLOG
  printf("XvMCCreateSurface: address=%x, offset=%x height=%d, width=%d\n",
	 pSavageSurface->data, pSavageSurface->offsets[0], surface->height, surface->width);
#endif

  /* Clear the surface to 0 */
  memset((void *)((unsigned int)pSavageSurface->data +
		  (unsigned int)pSavageSurface->offsets[0]),0,
	 (pSavageSurface->pitch * surface->height));

  pSavageXvMC->TileOffset = ((pSavageSurface->wWidth * 2 + 127) >> 7) << 8 ;

  pSavageXvMC->ref++;
#ifdef XVMCLOG
  printf("End XvMCCreateSurface\n");
#endif

  return Success;
}


/***************************************************************************
// Function: XvMCDestroySurface
***************************************************************************/
Status XvMCDestroySurface(Display *display, XvMCSurface *surface) {
  savageXvMCSurface *pSavageSurface;
  savageXvMCContext *pSavageXvMC;
#ifdef XVMCLOG
  printf("Start XvMCDestroySurface\n");
#endif
  if((display == NULL) || (surface == NULL)) {
    return BadValue;
  }
  if(surface->privData == NULL) {
    return (error_base + XvMCBadSurface);
  }

  pSavageSurface = (savageXvMCSurface *)surface->privData;
  if(pSavageSurface->last_flip) {
    XvMCSyncSurface(display,surface);
  }
  pSavageXvMC = (savageXvMCContext *)pSavageSurface->privContext;

  _xvmc_destroy_surface(display,surface);

  savage_free_privContext(pSavageXvMC);

  free(pSavageSurface);
  surface->privData = NULL;
#ifdef XVMCLOG
  printf("End XvMCDestroySurface\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCCreateBlocks
***************************************************************************/
Status XvMCCreateBlocks(Display *display, XvMCContext *context,
			unsigned int num_blocks,
			XvMCBlockArray *block) {

  savageXvMCContext *pSavageXvMC;
  unsigned int i;
  int found = -1;
#ifdef XVMCLOG
  printf("Start XvMCCreateBlock: num_blocks = %d\n", num_blocks);
#endif
  pSavageXvMC = (savageXvMCContext *)context->privData;

  if((display == NULL) || (context == NULL) || (num_blocks == 0)) {
    return BadValue;
  }

  for(i = 0; i < 4; i++)
  {
      if(!pSavageXvMC->DCTBlock[i].used)
      {
	  found = i;
	  pSavageXvMC->DCTBlock[i].used = 1;
	  pSavageXvMC->DCTBlock[i].dwTag = 0;
	  pSavageXvMC->DCTBlock[i].dwOldTag = 0;
	  break;
      }
  }

  if(found != -1)
      block->blocks = (short *)pSavageXvMC->DCTBlock[found].offset;
  else
      block->blocks = (short *)NULL;

  if(block->blocks == NULL) {
    return BadAlloc;
  }

  block->num_blocks = num_blocks;
  block->context_id = context->context_id;

  block->privData = pSavageXvMC;

#ifdef XVMCLOG
  printf("idct number=%d, idct offset=%x\n", found, (unsigned int)block->blocks);
#endif

  memset((void *)block->blocks,0,1024 * 1024L);
  pSavageXvMC->ref++;
#ifdef XVMCLOG
  printf("End XvMCCreateBlock\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCDestroyBlocks
***************************************************************************/
Status XvMCDestroyBlocks(Display *display, XvMCBlockArray *block) {
  savageXvMCContext *pSavageXvMC;
  unsigned int i;
#ifdef XVMCLOG
  printf("Start XvMCDestroyBlock\n");
#endif
  pSavageXvMC = (savageXvMCContext *)block->privData;

  if((display == NULL) || (pSavageXvMC == NULL)) {
    return BadValue;
  }

  for(i = 0; i < 4; i++)
  {
      if(pSavageXvMC->DCTBlock[i].offset == (unsigned int)block->blocks)
      {
	  pSavageXvMC->DCTBlock[i].used = 0;
	  break;
      }
  }

  block->num_blocks = 0;
  block->context_id = 0;
  block->privData = NULL;
  savage_free_privContext(pSavageXvMC);
#ifdef XVMCLOG
  printf("End XvMCDestroyBlock\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCCreateMacroBlocks
***************************************************************************/
Status XvMCCreateMacroBlocks(Display *display, XvMCContext *context,
			     unsigned int num_blocks,
			     XvMCMacroBlockArray *blocks) {
#ifdef XVMCLOG
  printf("Start XvMCCreateMacroBlocks: num_blocks = %d\n", num_blocks);
#endif
  if((display == NULL) || (context == NULL) || (blocks == NULL) ||
      (num_blocks == 0)) {
    return BadValue;
  }
  memset(blocks,0,sizeof(XvMCMacroBlockArray));
  blocks->context_id = context->context_id;
  blocks->privData = NULL;

  blocks->macro_blocks = (XvMCMacroBlock *)
    malloc(num_blocks * sizeof(XvMCMacroBlock));
  if(blocks->macro_blocks == NULL) {
    return BadAlloc;
  }
  blocks->num_blocks = num_blocks;
#ifdef XVMCLOG
  printf("End XvMCCreateMacroBlocks\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCDestroyMacroBlocks
***************************************************************************/
Status XvMCDestroyMacroBlocks(Display *display, XvMCMacroBlockArray *block) {
#ifdef XVMCLOG
  printf("Start XvMCDestroyMacroBlocks\n");
#endif
  if((display == NULL) || (block == NULL)) {
    return BadValue;
  }
  if(block->macro_blocks) {
    free(block->macro_blocks);
  }
  block->context_id = 0;
  block->num_blocks = 0;
  block->privData = NULL;
#ifdef XVMCLOG
  printf("Start XvMCDestroyMacroBlocks\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCRenderSurface
// Description: This function does the actual HWMC. Given a list of
//  macroblock structures it dispatched the hardware commands to execute
//  them. DMA buffer containing Y data are dispatched as they fill up
//  U and V DMA buffers are queued until all Y's are done. This minimizes
//  the context flipping and flushing required when switching between Y
//  U and V surfaces.
***************************************************************************/
Status XvMCRenderSurface(Display *display, XvMCContext *context,
			 unsigned int picture_structure,
			 XvMCSurface *target_surface,
			 XvMCSurface *past_surface,
			 XvMCSurface *future_surface,
			 unsigned int flags,
			 unsigned int num_macroblocks,
			 unsigned int first_macroblock,
			 XvMCMacroBlockArray *macroblock_array,
			 XvMCBlockArray *blocks) {
  /* Block Pointer */
  short *block_ptr;
  /* Current Macroblock Pointer */
  XvMCMacroBlock *mb;
  MCS3BeginData bd;
  int i;
  savageXvMCSurface *privTarget;
  savageXvMCSurface *privFuture = NULL;
  savageXvMCSurface *privPast = NULL;
  savageXvMCContext *pSavageXvMC;

  uint dwFrameDesc=0;

#ifdef XVMCLOG
  printf("Start XvMCRenderSurface\n");
#endif
  /* Check Parameters for validity */
  if((target_surface == NULL) || (context == NULL) || (display == NULL)) {
    printf("Error, Invalid Target,Context, or DIsplay!\n");
    return BadValue;
  }

  if(num_macroblocks == 0)
      {return Success;}

  if((macroblock_array == NULL) || (blocks == NULL))
      {return BadValue;}

  if(context->privData == NULL)
      {return BadValue;}

  pSavageXvMC = (savageXvMCContext *)context->privData;


  if(target_surface->privData == NULL) {
    printf("Error, Invalid Target Surface!\n");
    return BadValue;
  }
  privTarget = (savageXvMCSurface *)target_surface->privData;

  if(macroblock_array->num_blocks < (num_macroblocks + first_macroblock)) {
    printf("Error, Too many macroblocks requested for MB array size.\n");
    return BadValue;
  }

#ifdef XVMCLOG
  printf("width=%d, height=%d\n", target_surface->width, target_surface->height);
#endif

  /* Test For YV12 Surface */
  if(context->surface_type_id != FOURCC_YV12) {
    printf("Error, HWMC only possible on YV12 Surfaces\n");
    return BadValue;
  }

  if((past_surface == NULL) && (future_surface == NULL))
  {
      dwFrameDesc |= FRAME_I_PIC;
      pSavageXvMC->dwPictureType = FTYPE_I;
      privTarget->pType = FTYPE_I;
#ifdef XVMCLOG
      printf("FTYPE_I\n");
#endif
      privFuture = privTarget;
      privPast = privTarget;
  }
  else
  {

      if(future_surface == NULL)
      {
	  if(past_surface->privData == NULL) {
	      printf("Error, Invalid Past Surface!\n");
	      return BadValue;
	  }

#ifdef XVMCLOG
	  printf("FTYPE_P\n");
#endif
	  dwFrameDesc |= FRAME_P_PIC;
	  pSavageXvMC->dwPictureType = FTYPE_P;
	  privTarget->pType = FTYPE_P;
	  privPast = (savageXvMCSurface *)past_surface->privData;
	  privFuture = privTarget;

	  if((picture_structure != XVMC_FRAME_PICTURE) && !(flags ==  XVMC_SECOND_FIELD))
	  {
	      /* Savage Specific flag for marking second fields. */
	      dwFrameDesc |= FRAME_SECOND_P_FIELD;
	  }

	  /* return Success;*/	/* ---Bypass P-picture ---- */
      }
      else
      {
	  if((future_surface->privData == NULL) || (past_surface == NULL))
	  {
	      printf("Error, Invalid Future Surface or No Past Surface!\n");
	      return BadValue;
	  }
#ifdef XVMCLOG
	  printf("FTYPE_B\n");
#endif
	  dwFrameDesc |= FRAME_B_PIC;
	  pSavageXvMC->dwPictureType = FTYPE_B;
	  privTarget->pType = FTYPE_B;
	  privPast = (savageXvMCSurface *)past_surface->privData;  /* Added By Giga Hsu */
	  privFuture = (savageXvMCSurface *)future_surface->privData;

	  /* return Success; */	 /* ---Bypass B-picture ---- */
      }
  }

  switch(picture_structure)
  {
      case XVMC_TOP_FIELD:
	   dwFrameDesc |= FRAME_TOP_FIELD;
#ifdef XVMCLOG
	   printf("top field\n");
#endif
	   break;

      case XVMC_BOTTOM_FIELD:
	   dwFrameDesc |= FRAME_BOTTOM_FIELD;
#ifdef XVMCLOG
	   printf("bottom field\n");
#endif
	   break;

      case XVMC_FRAME_PICTURE:
	   dwFrameDesc |= FRAME_FRAME;
#ifdef XVMCLOG
	   printf("frame\n");
#endif
	   break;
  }

  mb = &macroblock_array->macro_blocks[first_macroblock];
  if(mb->motion_type==0)
  {
  /*	mb->motion_type=2;*/
  }

  pSavageXvMC->lpIDCTBuff = (unsigned char *)&(blocks->blocks[mb->index << 6]);

#ifdef XVMC
  printf("idct buffer=%x\n", pSavageXvMC->lpIDCTBuff);
#endif
  /* Convert to macroblock based width and height */
  bd.width = ((privTarget->wWidth + 63) & 0xffffffc0) >> 4;
  bd.height = privTarget->wHeight >> 4;
  bd.dwFrameDesc = dwFrameDesc;
  MCS3BeginPicture(pSavageXvMC, &bd);

#ifdef XVMCLOG
  printf("first_macroblock=%d;num_macroblocks=%d\n",first_macroblock,num_macroblocks);
#endif

  if(privTarget->pType == FTYPE_P)
	framenum++;
  for(i=first_macroblock; i<(num_macroblocks + first_macroblock); i++)
  {
     unsigned int dwMacroBlockDesc = 0;
     unsigned int dwMacroMask = 0;
     MCS3MacroBLKData MB;
     unsigned int mbType, mbTypeWasNot_0,MotionType,motion_vertical_field_select;
     /* Set up values needed for each macroblock */
     unsigned int iV;
     BOOL bPicCurrentField =  1 - (picture_structure & 2 >> 1);	   /*  Fields only */
     for(iV = 0; iV < 4; iV++)
     {
	 MB.wMV[iV] = 0;
	 MB.wMV[iV+4] = 0;
     }

     mb = &macroblock_array->macro_blocks[i];
  if(privTarget->pType == FTYPE_P)
  {
	if(framenum==1)
	{
#ifdef XVMCLOG
	  printf("framenum=%d,x=%d,y=%d,mbtype=%x,motion_type=%x,pattern=%x,PMV=%d,%d,%d,%d\n",
		 framenum,mb->x,mb->y,mb->macroblock_type,mb->motion_type,mb->coded_block_pattern,
		 mb->PMV[0][0][0],mb->PMV[0][0][1],mb->PMV[1][0][0],mb->PMV[1][0][1]);
#endif
	}
  }

  if(mb->motion_type==0)
  {
	/*mb->motion_type=2;*/
  }

     block_ptr = &(blocks->blocks[mb->index<<5]);

    /* Lockup can happen if the coordinates are too far out of range */
    if(mb->x > target_surface->width>>4) {
      mb->x = 0;
    }

    if(mb->y > target_surface->height>>4) {
      mb->y = 0;
    }

    /* printf("h=%d, v=%d\n", mb->x, mb->y); */

    MB.dwMBLKHOff = mb->x << 4;
    MB.dwMBLKVOff = mb->y << 4;
    MB.lpIDCTBuff = (void *)block_ptr;

    if(mb->dct_type == XVMC_DCT_TYPE_FIELD)
    {
	dwMacroBlockDesc |= MB_DCT_FIELD;
#ifdef XVMCLOG
	printf("DCT_FIELD\n");
#endif
    }

#ifdef XVMCLOG
    printf("coded_pattern=%x\n", mb->coded_block_pattern);
#endif

    if(mb->macroblock_type & XVMC_MB_TYPE_INTRA)
    {
	dwMacroBlockDesc |= MB_DECODE_CrBLK |
			    MB_DECODE_CbBLK |
			    MB_DECODE_Y3BLK |
			    MB_DECODE_Y2BLK |
			    MB_DECODE_Y1BLK |
			    MB_DECODE_Y0BLK;
#ifdef XVMCLOG
	 printf("Intra block\n");
#endif
	 pSavageXvMC->dwOverFlow=0;   /* Added by Giga */

	 if(mb->coded_block_pattern & 0x8000)
	      pSavageXvMC->dwOverFlow=2;   /* Added by John Zhao */

    }
    else
    {
	pSavageXvMC->dwOverFlow=1;   /* Added by Giga */

	mbType = mb->macroblock_type; /* change sequence by Giga Hsu */
	mbType = (mbType >> 1) & 0x3;

	if(mbType==0)
	{
	    mbType = 1;
	    mbTypeWasNot_0 = FALSE;
	}
	else
	{
	    mbTypeWasNot_0 = TRUE;
	}

#ifdef XVMCLOG
	printf("coded_block_pattern=%x, mbType=%x, motion_type=%x\n", mb->coded_block_pattern, mbType, mb->motion_type);
#endif
	MotionType=mb->motion_type;
	dwMacroBlockDesc |= mbType << 7;
	dwMacroBlockDesc |= MotionType << 9;
	dwMacroBlockDesc |= mb->coded_block_pattern & 0x3f;

	/* if it is intrablock, we will set all the block with overflow. */

	motion_vertical_field_select=mb->motion_vertical_field_select;

	dwMacroBlockDesc |= (unsigned int)(mb->coded_block_pattern & 0x3f) << 16;

	if(mb->coded_block_pattern & 0x8000)
	      pSavageXvMC->dwOverFlow=2;   /* Added by John Zhao */

	switch(picture_structure)
	{
	    case XVMC_TOP_FIELD:
	    case XVMC_BOTTOM_FIELD:
#if  0
		 printf("Field picture\n");
#endif
		 switch(mb->motion_type)
		 {
		     case XVMC_PREDICTION_16x8:	 /* Frame Motion */
			  switch(mbType)
			  {
			      case 1:
				   dwMacroMask = MB_MV0_ENABLED | MB_MV1_ENABLED;
				   if(mbTypeWasNot_0)
				   {
				       MB.wMV[0] = mb->PMV[0][0][0];
				       MB.wMV[1] = mb->PMV[0][0][1] & 0x0efff;
				       MB.wMV[2] = mb->PMV[1][0][0];
				       MB.wMV[3] = mb->PMV[1][0][1] & 0x0efff;
				   }
				   break;

			      case 2:
				   dwMacroMask = MB_MV2_ENABLED | MB_MV3_ENABLED;
				   MB.wMV[4] = mb->PMV[0][1][0];
				   MB.wMV[5] = mb->PMV[0][1][1] & 0x0efff;
				   MB.wMV[6] = mb->PMV[1][1][0];
				   MB.wMV[7] = mb->PMV[1][1][1] & 0x0efff;
				   break;

			      case 3:
				   dwMacroMask = MB_MV0_ENABLED |  MB_MV1_ENABLED |
						 MB_MV2_ENABLED |  MB_MV3_ENABLED;
				   MB.wMV[0] = mb->PMV[0][0][0];
				   MB.wMV[1] = mb->PMV[0][0][1] & 0x0efff;
				   MB.wMV[2] = mb->PMV[1][0][0];
				   MB.wMV[3] = mb->PMV[1][0][1] & 0x0efff;
				   MB.wMV[4] = mb->PMV[0][1][0];
				   MB.wMV[5] = mb->PMV[0][1][1] & 0x0efff;
				   MB.wMV[6] = mb->PMV[1][1][0];
				   MB.wMV[7] = mb->PMV[1][1][1] & 0x0efff;
				   break;
			  }
			  break;

		     case XVMC_PREDICTION_FIELD: /* Field Motion */
			  switch(mbType)
			  {
			      case 1:
				   dwMacroMask = MB_MV0_ENABLED;
				   if(mbTypeWasNot_0)
				   {
				       MB.wMV[0] = mb->PMV[0][0][0];
				       MB.wMV[1] = mb->PMV[0][0][1] & 0x0efff;
				       MB.wMV[1] |= ((motion_vertical_field_select &
						     XVMC_SELECT_FIRST_FORWARD) << 12);
				   }
				   else
				   {
				       MB.wMV[1] |= (bPicCurrentField) << 12;
				   }
				   break;

			      case 2:
				   dwMacroMask = MB_MV2_ENABLED;
				   MB.wMV[4] = mb->PMV[0][1][0];
				   MB.wMV[5] = mb->PMV[0][1][1] & 0x0efff;
				   MB.wMV[5] |= ((motion_vertical_field_select &
						 XVMC_SELECT_FIRST_BACKWARD) << 11);
				   break;

			      case 3:
				   dwMacroMask = MB_MV0_ENABLED |  MB_MV2_ENABLED;
				   MB.wMV[0] = mb->PMV[0][0][0];
				   MB.wMV[1] = mb->PMV[0][0][1] & 0x0efff;
				   MB.wMV[1] |= ((motion_vertical_field_select &
						XVMC_SELECT_FIRST_FORWARD) << 12);
				   MB.wMV[4] = mb->PMV[0][1][0];
				   MB.wMV[5] = mb->PMV[0][1][1] & 0x0efff;
				   MB.wMV[5] |= ((motion_vertical_field_select &
						 XVMC_SELECT_FIRST_BACKWARD) << 11);
				   break;
			  }
			  break;

		     case XVMC_PREDICTION_DUAL_PRIME:
			    dwMacroMask = MB_MV0_ENABLED |  MB_MV1_ENABLED;
			  /* if(mbTypeWasNot_0) */
			  {
			    MB.wMV[0] = mb->PMV[0][0][0];
			    MB.wMV[1] = mb->PMV[0][0][1] & 0x0efff;
			    MB.wMV[1] |=  (bPicCurrentField) << 12;

			    MB.wMV[2] = mb->PMV[0][1][0];
			    MB.wMV[3] = mb->PMV[0][1][1] & 0x0efff;
			    MB.wMV[3] |=(1 - bPicCurrentField) << 12;
			  }
			  break;
		 }
		 break;

	    case XVMC_FRAME_PICTURE:
#if  0
		 printf("Frame Picture\n");
#endif
		 switch(mb->motion_type)
		 {
		     case XVMC_PREDICTION_FIELD:   /* Field motion */
			  switch(mbType)
			  {
			      case 1:
#ifdef XVMCLOG
				   printf("Field motion1\n");
#endif
				   dwMacroMask = MB_MV0_ENABLED | MB_MV1_ENABLED;
				   if(mbTypeWasNot_0)
				   {
				       MB.wMV[0] =  mb->PMV[0][0][0];
				       MB.wMV[1] = (mb->PMV[0][0][1] >> 1) & 0x0efff;
				       MB.wMV[1] |= ((motion_vertical_field_select &
						     XVMC_SELECT_FIRST_FORWARD) << 12);
				       MB.wMV[2] =  mb->PMV[1][0][0];
				       MB.wMV[3] = (mb->PMV[1][0][1] >> 1) & 0x0efff;
				       MB.wMV[3] |= ((motion_vertical_field_select &
						     XVMC_SELECT_SECOND_FORWARD) << 10);
				   }
				   break;

			      case 2:
#ifdef XVMCLOG
				   printf("Field motion2\n");
#endif
				   dwMacroMask = MB_MV2_ENABLED | MB_MV3_ENABLED;
				   MB.wMV[4] =	mb->PMV[0][1][0];
				   MB.wMV[5] = (mb->PMV[0][1][1] >> 1) & 0x0efff;
				   MB.wMV[5] |= ((motion_vertical_field_select &
						 XVMC_SELECT_FIRST_BACKWARD) << 11);
				   MB.wMV[6] =	mb->PMV[1][1][0];
				   MB.wMV[7] = (mb->PMV[1][1][1] >> 1) & 0x0efff;
				   MB.wMV[7] |= ((motion_vertical_field_select &
						 XVMC_SELECT_SECOND_BACKWARD) << 9);
				   break;

			      case 3:
#ifdef XVMCLOG
				   printf("Field motion3\n");
#endif
				   dwMacroMask = MB_MV0_ENABLED |  MB_MV1_ENABLED |
						 MB_MV2_ENABLED |  MB_MV3_ENABLED;
				   MB.wMV[0] =	mb->PMV[0][0][0];
				   MB.wMV[1] = (mb->PMV[0][0][1] >> 1) & 0x0efff;
				   MB.wMV[1] |= ((motion_vertical_field_select &
						 XVMC_SELECT_FIRST_FORWARD)  << 12);
				   MB.wMV[2] =	mb->PMV[1][0][0];
				   MB.wMV[3] = (mb->PMV[1][0][1] >> 1) & 0x0efff;
				   MB.wMV[3] |= ((motion_vertical_field_select &
						 XVMC_SELECT_SECOND_FORWARD) << 10);
				   MB.wMV[4] =	mb->PMV[0][1][0];
				   MB.wMV[5] = (mb->PMV[0][1][1] >> 1) & 0x0efff;
				   MB.wMV[5] |= ((motion_vertical_field_select &
						 XVMC_SELECT_FIRST_BACKWARD) << 11);
				   MB.wMV[6] =	mb->PMV[1][1][0];
				   MB.wMV[7] = (mb->PMV[1][1][1] >> 1) & 0x0efff;
				   MB.wMV[7] |= ((motion_vertical_field_select &
						 XVMC_SELECT_SECOND_BACKWARD)<< 9);
				   break;
			  }
			  break;

		     case XVMC_PREDICTION_FRAME:  /* Frame (16x8 MC) */

			  switch(mbType)
			  {
			      case 1:
				   dwMacroMask = MB_MV0_ENABLED;
				   if(mbTypeWasNot_0)
				   {
				       MB.wMV[0] = mb->PMV[0][0][0];
				       MB.wMV[1] = mb->PMV[0][0][1];
				   }
				   break;

			      case 2:
				   dwMacroMask = MB_MV2_ENABLED;
				   MB.wMV[4] = mb->PMV[0][1][0];
				   MB.wMV[5] = mb->PMV[0][1][1];
				   break;

			      case 3:
				   dwMacroMask = MB_MV0_ENABLED | MB_MV2_ENABLED;
				   MB.wMV[0] = mb->PMV[0][0][0];
				   MB.wMV[1] = mb->PMV[0][0][1];
				   MB.wMV[4] = mb->PMV[0][1][0];
				   MB.wMV[5] = mb->PMV[0][1][1];
				   break;
			  }
			  break;

		     case XVMC_PREDICTION_DUAL_PRIME:
			   dwMacroMask = MB_MV0_ENABLED |  MB_MV1_ENABLED |
					 MB_MV2_ENABLED |  MB_MV3_ENABLED ;
			  /* if(mbTypeWasNot_0) */
			  {
			    MB.wMV[0] = mb->PMV[0][0][0];
			    MB.wMV[1] = mb->PMV[0][0][1] & 0x0efff;
			    MB.wMV[1] |= (0 << 12);

			    MB.wMV[2] = mb->PMV[0][1][0];
			    MB.wMV[3] = mb->PMV[0][1][1] & 0x0efff;
			    MB.wMV[3] |= (1 << 12);

			    MB.wMV[4] = mb->PMV[1][0][0];
			    MB.wMV[5] = mb->PMV[1][0][1] & 0x0efff;
			    MB.wMV[5] |= (1 << 12);

			    MB.wMV[6] = mb->PMV[1][1][0];
			    MB.wMV[7] = mb->PMV[1][1][1] & 0x0efff;
			    MB.wMV[7] |= (0 << 12);
			  }
			  break;
		 }
		 break;
	}
	dwMacroBlockDesc |= dwMacroMask;
    }
    MB.dwMacroBLKDesc = dwMacroBlockDesc;
    MCS3MacroBlock(&MB, pSavageXvMC);
  } /* for each Macroblock */

  MCS3EndPicture(pSavageXvMC);

  /* SAVAGE_LOCK(pSavageXvMC,0); Moved into MCS3DecodeFrame(). */
  MCS3DecodeFrame(pSavageXvMC, privTarget, privFuture, privPast);
  /* SAVAGE_UNLOCK(pSavageXvMC); Moved into MCS3DecodeFrame(). */

#ifdef XVMCLOG
  printf("End XvMCRenderSurface\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCPutSurface
// Description:
// Arguments:
//  display: Connection to X server
//  surface: Surface to be displayed
//  draw: X Drawable on which to display the surface
//  srcx: X coordinate of the top left corner of the region to be
//	    displayed within the surface.
//  srcy: Y coordinate of the top left corner of the region to be
//	    displayed within the surface.
//  srcw: Width of the region to be displayed.
//  srch: Height of the region to be displayed.
//  destx: X cordinate of the top left corner of the destination region
//	   in the drawable coordinates.
//  desty: Y cordinate of the top left corner of the destination region
//	   in the drawable coordinates.
//  destw: Width of the destination region.
//  desth: Height of the destination region.
//  flags: One or more of the following.
//     XVMC_TOP_FIELD - Display only the Top field of the surface.
//     XVMC_BOTTOM_FIELD - Display only the Bottom Field of the surface.
//     XVMC_FRAME_PICTURE - Display both fields or frame.
//
// Info: Portions of this function derived from savage_video.c (XFree86)
//
//   This function is organized so that we wait as long as possible before
//   touching the overlay registers. Since we don't know that the last
//   flip has happened yet we want to give the overlay as long as
//   possible to catch up before we have to check on its progress. This
//   makes it unlikely that we have to wait on the last flip.
***************************************************************************/
Status XvMCPutSurface(Display *display,XvMCSurface *surface,
		      Drawable draw, short srcx, short srcy,
		      unsigned short srcw, unsigned short srch,
		      short destx, short desty,
		      unsigned short destw, unsigned short desth,
		      int flags) {
  savageXvMCContext *pSavageXvMC;
  savageXvMCSurface *pSavageSurface;
  Box extents;
  uint window_width,window_height;
  uint d;
  double xscale,yscale;
  int diff;
  int clipped_srcx, clipped_srcy, clipped_destx, clipped_desty;
  int clipped_srcw, clipped_srch, clipped_destw, clipped_desth;
  int x1,y1;  /* KGy Left_Out */
  uint root_width,root_height;
  int x2 = 0, y2 = 0,unused;
  uint nChilds;
  Window win,root,parent,*pChilds;
#ifdef XVMCLOG
  unsigned int width, height;
#endif
  unsigned int isChanged=0;
  GC   gc;
#ifdef XVMCLOG
  XvImage * image;
#endif
  char * data;

#ifdef XVMCLOG
  printf("Start XvMCPutSurface: dest=%d, desty=%d, destw=%d, desth=%d\n", destx, desty, destw, desth);
#endif
  if((display == NULL) || (surface == NULL)) {
    return BadValue;
  }

  if(surface->privData == NULL) {
    return (error_base + XvMCBadSurface);
  }
  pSavageSurface = (savageXvMCSurface *)surface->privData;
  pSavageXvMC = (savageXvMCContext *)pSavageSurface->privContext;
#if 0
  pORegs = (savageOverlayRecPtr)pSavageXvMC->oregs;
#endif

#if 1

  /*
    FIXME: This should be using the DRI's clip rect but that isn't
    all hooked up yet. This has some latency but we get by.
  */

  XLockDisplay(display);
  win = draw;
  XQueryTree(display,win,&root,&parent,&pChilds,&nChilds);
  if(nChilds) XFree(pChilds);
  XGetGeometry(display,win, &root, &x2, &y2, &window_width,
	       &window_height, &d, &d);
  XUnlockDisplay(display);
  x1 = x2;
  y1 = y2;
  win = parent;
  while(win != root) {
    XLockDisplay(display);
    XQueryTree(display,win,&root,&parent,&pChilds,&nChilds);
    if(nChilds) XFree(pChilds);
    XGetGeometry(display,win, &root, &x2, &y2, &d, &d, &d, &d);
    XUnlockDisplay(display);
    x1 += x2;
    y1 += y2;
    win = parent;
  };

  XLockDisplay(display);
  XGetGeometry(display,root, &root, &unused, &unused,
	       &root_width, &root_height, &d, &d);

  XUnlockDisplay(display);

#ifdef XVMCLOG
  printf("PutSurface1: destx=%d, desty=%d, x1=%d, y1=%d\n", destx, desty, x1, y1);
  printf("PutSurface1: srcx=%d, srcyy=%d, srcw=%d, srch=%d\n", srcx, srcy, srcw, srch);
#endif

  /* Left edge of Video window clipped to screen */
  extents.x1 = 0;
  if(x1 > extents.x1) {
    extents.x1 = x1;
  }
  /* Right edge of Video window clipped to screen */
  extents.x2 = root_width;
  if(extents.x2 > (x1 + window_width)) {
    extents.x2 = x1 + window_width;
  }
  /* Top edge of Video window clipped to screen */
  extents.y1 = 0;
  if(y1 > extents.y1) {
    extents.y1 = y1;
  }
  /* Bottom edge of Video window clipped to screen */
  extents.y2 = root_height;
  if(extents.y2 > (y1 + window_height)) {
    extents.y2 = y1 + window_height;
  }

  /*
    Clipping is more difficult than is seems. We need to keep the
    scaling factors even if the destination window needs to be clipped.
    We clip the destination window first then apply a scaled version
    to the source window.
  */
  /* Put destination coords in screen coords */
#ifdef XVMCLOG
  printf("PutSurface: destx=%d, desty=%d, x1=%d, y1=%d\n", destx, desty, x1, y1);
#endif
  destx += x1;
  desty += y1;

  /* Scale factors requested */
  xscale = (double)srcw / (double)destw;
  yscale = (double)srch / (double)desth;

  /*
    If destination window needs to be clipped we actually adjust both
    the src and dest window so as to keep the scaling that was requested
  */
  clipped_srcx = srcx;
  clipped_srcy = srcy;
  clipped_destx = destx;
  clipped_desty = desty;
  clipped_srcw = srcw;
  clipped_srch = srch;
  clipped_destw = destw;
  clipped_desth = desth;
  /* Clip to the source surface boundaries */
  if(clipped_srcx < 0) {
    clipped_destx += (0 - clipped_srcx) / xscale;
    clipped_srcw += clipped_srcx;
    clipped_destw += clipped_srcx / xscale;
    clipped_srcx = 0;
  }
  if((clipped_srcw + clipped_srcx) > surface->width) {
    clipped_srcw = surface->width - clipped_srcx;
    clipped_destw -= (clipped_srcw - srcw) / xscale;
  }
  if(clipped_srcy < 0) {
    clipped_desty += (0 - clipped_srcy) / yscale;
    clipped_srch += clipped_srcy;
    clipped_desth += clipped_srcy / yscale;
    clipped_srcy = 0;
  }
  if((clipped_srch + clipped_srcy) > surface->height) {
    clipped_srch = surface->height - clipped_srcy;
    clipped_desth -= (clipped_srch - srch) / yscale;
  }

  /* Clip to the extents */
  if(clipped_destx < extents.x1) {
    diff = extents.x1 - clipped_destx;
    clipped_srcx += diff * xscale;
    clipped_srcw -= diff * xscale;
    clipped_destw -= diff;
    clipped_destx = extents.x1;
  }

  diff = (clipped_destx + clipped_destw) - extents.x2;
  if(diff > 0) {
    clipped_destw -= diff;
    clipped_srcw -= diff * xscale;
  }

  if(clipped_desty < extents.y1) {
    diff = extents.y1 - clipped_desty;
    clipped_srcy += diff * yscale;
    clipped_srch -= diff * yscale;
    clipped_desth -= diff;
    clipped_desty = 0;
  }

  diff = (clipped_desty + clipped_desth) - extents.y2;
  if(diff > 0) {
    clipped_desth -= diff;
    clipped_srch -= diff * yscale;
  }

  /* If the whole window is clipped turn off the overlay */
  if((clipped_destx + clipped_destw < extents.x1) ||
     (clipped_desty + clipped_desth < extents.y1) ||
     (clipped_destx > extents.x2) ||
     (clipped_desty > extents.y2)) {
    return XvMCHideSurface(display, surface);
  }

#endif

#if 1
  if((pSavageXvMC->Overlay.srcx != clipped_srcx) ||
     (pSavageXvMC->Overlay.srcy != clipped_srcy) ||
     (pSavageXvMC->Overlay.destx != clipped_destx) ||
     (pSavageXvMC->Overlay.desty != clipped_desty) ||
     (pSavageXvMC->Overlay.srcWidth != clipped_srcw) ||
     (pSavageXvMC->Overlay.srcHeight != clipped_srch) ||
     (pSavageXvMC->Overlay.destWidth != clipped_destw) ||
     (pSavageXvMC->Overlay.destHeight != clipped_desth) ||
     (pSavageXvMC->dwFlags & COLORKEYCHANGED))
  {
#ifdef XVMCLOG
      printf("Changed: destwidth=%d, destheight=%d, srcwidth=%d, srcheight=%d\n", clipped_destw, clipped_desth, clipped_srcw, clipped_srch);
      printf("Changed: x=%d, y=%d\n", clipped_destx, clipped_desty);
#endif

      isChanged = 1;

      /* Colorkey is handled in setup video overlay. No separated handling, so reset the flag here. */
      pSavageXvMC->dwFlags &= ~COLORKEYCHANGED;
  }
#endif
  pSavageXvMC->Overlay.srcx = clipped_srcx;
  pSavageXvMC->Overlay.srcy = clipped_srcy;
  pSavageXvMC->Overlay.destx = clipped_destx;
  pSavageXvMC->Overlay.desty = clipped_desty;

  pSavageXvMC->Overlay.srcWidth = clipped_srcw;
  pSavageXvMC->Overlay.srcHeight = clipped_srch;
  pSavageXvMC->Overlay.destWidth = clipped_destw;
  pSavageXvMC->Overlay.destHeight = clipped_desth;

  /* For scale factor precision. */
  pSavageXvMC->Overlay.sw = srcw;
  pSavageXvMC->Overlay.dw = destw;

  pSavageSurface->srcx = clipped_srcx;
  pSavageSurface->srcy = clipped_srcy;
  pSavageSurface->destx = clipped_destx;
  pSavageSurface->desty = clipped_desty;

  pSavageSurface->srcw = clipped_srcw;
  pSavageSurface->srch = clipped_srch;
  pSavageSurface->destw = clipped_destw;
  pSavageSurface->desth = clipped_desth;

  if(flags == XVMC_FRAME_PICTURE)
  {
#ifdef XVMCLOG
      printf("PutSurface:display frame\n");
#endif
      pSavageSurface->dwDisplayTaskType = DISPLAY_FRAME;
  }
  else
  if(flags == XVMC_TOP_FIELD)
  {
#ifdef XVMCLOG
      printf("PutSurface:top field\n");
#endif
      pSavageSurface->dwDisplayTaskType = DISPLAY_TOP_FIELD;
  } else if(flags == XVMC_BOTTOM_FIELD) {
      pSavageSurface->dwDisplayTaskType = DISPLAY_BOTTOM_FIELD;
#ifdef XVMCLOG
      printf("PutSurface:bottom field\n");
#endif
  }

  if(clipped_srch > 1)
      clipped_srch--;

  if( xscale > 1 ) {
    pSavageSurface->destw = (clipped_destw * pSavageSurface->wWidth) / clipped_srcw;
    pSavageSurface->srcw = pSavageSurface->wWidth;
  }

  if(clipped_srch > clipped_desth)
  {
      pSavageSurface->desth = (clipped_desth * pSavageSurface->wHeight) / clipped_srch;
      pSavageSurface->srch = pSavageSurface->wHeight;
  }
  else
  {
      pSavageSurface->desth = clipped_desth;
      pSavageSurface->srch = clipped_srch;
  }
#ifdef XVMCLOG
  printf("SurfWidth=%d, SurfH=%d\n", pSavageSurface->wWidth,pSavageSurface->wHeight);
#endif

#ifdef XVMCLOG
  if(pSavageSurface->pType == FTYPE_I)
#endif
  {
#ifdef XVMCOG
     printf("PutSurface: Locking savage.\n");
#endif
     SAVAGE_LOCK(pSavageXvMC,0);
#ifdef XVMCOG
     printf("PutSurface: Locked savage.\n");
#endif
     if(isChanged)
	 MCS3SetupVideoOverlay(pSavageXvMC, pSavageSurface);

     MCS3DisplayFrame(pSavageXvMC, pSavageSurface);
#ifdef XVMCOG
     printf("PutSurface: Unlocking savage.\n");
#endif
     SAVAGE_UNLOCK(pSavageXvMC);
#ifdef XVMCOG
     printf("PutSurface: Unlocked savage.\n");
#endif
  }
  if(pSavageXvMC->dwFlags | STREAM_ON)
      pSavageXvMC->dwFlags = STREAM_ON;
  else
      pSavageXvMC->dwFlags = 0;

#ifdef XVMCLOG
  printf("End XvMCPutSurface\n");
#endif

  return Success;
}

/***************************************************************************
// Function: XvMCSyncSurface
// Arguments:
//   display - Connection to the X server
//   surface - The surface to synchronize
// Info:
// Returns: Status
***************************************************************************/
Status XvMCSyncSurface(Display *display,XvMCSurface *surface) {
  Status ret;
  int stat=0;
  /*
    FIXME: Perhaps a timer here to prevent lockup?
    FIXME: Perhaps a usleep to not be busy waiting?
  */
#ifdef XVMCLOG
  printf("Start sync surface \n");
#endif

  do {
    ret = XvMCGetSurfaceStatus(display,surface,&stat);
  }while(!ret && (stat & XVMC_RENDERING));
#ifdef XVMCLOG
  printf("End sync surface\n");
#endif
  return ret;
}

/***************************************************************************
// Function: XvMCFlushSurface
// Description:
//   This function commits pending rendering requests to ensure that they
//   wll be completed in a finite amount of time.
// Arguments:
//   display - Connection to X server
//   surface - Surface to flush
// Info:
//   This command is a noop for savage becuase we always dispatch buffers in
//   render. There is little gain to be had with 4k buffers.
// Returns: Status
***************************************************************************/
Status XvMCFlushSurface(Display * display, XvMCSurface *surface) {
  return Success;
}

/***************************************************************************
// Function: XvMCGetSurfaceStatus
// Description:
// Arguments:
//  display: connection to X server
//  surface: The surface to query
//  stat: One of the Following
//    XVMC_RENDERING - The last XvMCRenderSurface command has not
//		       completed.
//    XVMC_DISPLAYING - The surface is currently being displayed or a
//		       display is pending.
***************************************************************************/
Status XvMCGetSurfaceStatus(Display *display, XvMCSurface *surface,
			    int *stat) {
  savageXvMCSurface *privSurface;
  savageXvMCContext *pSavageXvMC;
  int temp;

  if((display == NULL) || (surface == NULL) || (stat == NULL)) {
    return BadValue;
  }
  if(surface->privData == NULL) {
    return BadValue;
  }
  *stat = 0;
  privSurface = surface->privData;

  pSavageXvMC = privSurface->privContext;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadSurface);
  }

#if 0
/*  SAVAGE_LOCK(pSavageXvMC,0);*/
  if(privSurface->last_flip) {
    /* This can not happen */
    if(pSavageXvMC->last_flip < privSurface->last_flip) {
      printf("Error: Context last flip is less than surface last flip.\n");
      return BadValue;
    }
    /*
      If the context has 2 or more flips after this surface it
      cannot be displaying. Don't bother to check.
    */
    if(!(pSavageXvMC->last_flip > (privSurface->last_flip + 1))) {
      /*
	If this surface was the last flipped it is either displaying
	or about to be so don't bother checking.
      */
      if(pSavageXvMC->last_flip == privSurface->last_flip) {
	*stat |= XVMC_DISPLAYING;
      }
      else {
	/*
	  In this case there has been one more flip since our surface's
	  but we need to check if it is finished or not.
	*/
	temp = GET_FSTATUS(pSavageXvMC);
	if(((temp & (1<<20))>>20) != pSavageXvMC->current) {
	  *stat |= XVMC_DISPLAYING;
	}
      }
    }
  }

  if(privSurface->last_render &&
      (privSurface->last_render > GET_RSTATUS(pSavageXvMC))) {
    *stat |= XVMC_RENDERING;
  }
/*  SAVAGE_UNLOCK(pSavageXvMC);*/
#endif

  return Success;
}

/***************************************************************************
//
//  Surface manipulation functions
//
***************************************************************************/

/***************************************************************************
// Function: XvMCHideSurface
// Description: Stops the display of a surface.
// Arguments:
//   display - Connection to the X server.
//   surface - surface to be hidden.
//
// Returns: Status
***************************************************************************/
Status XvMCHideSurface(Display *display, XvMCSurface *surface) {
  savageXvMCSurface *pSavageSurface;
  savageXvMCContext *pSavageXvMC;
  int ss, xx;

#ifdef XVMCLOG
  printf("Start XvMCHideSurface\n");
#endif
  /* Did we get a good display and surface passed into us? */
  if(display == NULL) {
    return BadValue;
  }

  if(surface == NULL) {
    return (error_base + XvMCBadSurface);
  }

  XvMCSyncSurface(display, surface);

  /* Get surface private data pointer */
  if(surface->privData == NULL) {
    return (error_base + XvMCBadSurface);
  }
  pSavageSurface = (savageXvMCSurface *)surface->privData;

  /*
    Get the status of the surface, if it is not currently displayed
    we don't need to worry about it.
  */
  if((xx = XvMCGetSurfaceStatus(display, surface, &ss)) != Success) {
    return xx;
  }
  if(! (ss & XVMC_DISPLAYING)) {
    return Success;
  }

  /* Get the associated context pointer */
  pSavageXvMC = (savageXvMCContext *)pSavageSurface->privContext;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadSurface);
  }

  if(pSavageXvMC->last_flip) {

#if 0
    /* Make sure last flip is done */
    BLOCK_OVERLAY(pSavageXvMC,pSavageXvMC->current);

    /* Set the registers to turn the overlay off */
    pSavageXvMC->oregs->OV0CMD = VC_UP_INTERPOLATION | HC_UP_INTERPOLATION |
      Y_ADJUST;
    pSavageXvMC->current = !pSavageXvMC->current;
    if(pSavageXvMC->current == 1) {
      pSavageXvMC->oregs->OV0CMD |= BUFFER1_FIELD0;
    }
    else {
      pSavageXvMC->oregs->OV0CMD |= BUFFER0_FIELD0;
    }
    OVERLAY_FLIP(pSavageXvMC);
    /*
      Increment the context flip but not the surface. This way no
      surface has the last flip #.
    */
    pSavageXvMC->last_flip++;


    /* Now wait until the hardware reads the registers and makes the change. */
    BLOCK_OVERLAY(pSavageXvMC,pSavageXvMC->current)
#endif

  }
#ifdef XVMCLOG
  printf("End XvMCHideSurface\n");
#endif
  return Success;
}




/***************************************************************************
//
// Functions that deal with subpictures

//
***************************************************************************/



/***************************************************************************
// Function: XvMCCreateSubpicture
// Description: This creates a subpicture by filling out the XvMCSubpicture
//		structure passed to it and returning Success.
// Arguments:
//   display - Connection to the X server.
//   context - The context to create the subpicture for.
//   subpicture - Pre-allocated XvMCSubpicture structure to be filled in.
//   width - of subpicture
//   height - of subpicture
//   xvimage_id - The id describing the XvImage format.
//
// Returns: Status
***************************************************************************/
Status XvMCCreateSubpicture(Display *display, XvMCContext *context,
			  XvMCSubpicture *subpicture,
			  unsigned short width, unsigned short height,
			  int xvimage_id) {
  savageXvMCContext *pSavageXvMC;
  savageXvMCSubpicture *pSavageSubpicture;

#ifdef XVMCLOG
  printf("Start XvMCCreateSubpicture\n");
#endif

  if((subpicture == NULL) || (context == NULL) || (display == NULL)){
    return BadValue;
  }

  pSavageXvMC = (savageXvMCContext *)context->privData;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadContext);
  }


  subpicture->context_id = context->context_id;
  subpicture->xvimage_id = xvimage_id;

  /* These need to be checked to make sure they are not too big! */
  subpicture->width = width;
  subpicture->height = height;

  subpicture->privData =
    (savageXvMCSubpicture *)malloc(sizeof(savageXvMCSubpicture));

  if(!subpicture->privData) {
    return BadAlloc;
  }
  pSavageSubpicture = (savageXvMCSubpicture *)subpicture->privData;


  /* Data == Client Address, offset == Physical address offset */
  pSavageSubpicture->data = 0;
  pSavageSubpicture->handle = 0;

  /* Initialize private values */
  pSavageSubpicture->privContext = pSavageXvMC;

  pSavageSubpicture->last_render = 0;
  pSavageSubpicture->last_flip = 0;

  /* Based on the xvimage_id we will need to set the other values */
  subpicture->num_palette_entries = 16;
  subpicture->entry_bytes = 3;
  strcpy(subpicture->component_order,"YUV");

  /*
    savage's MC Engine needs surfaces of 2^x (x= 9,10,11,12) pitch
    and the Tiler need 512k aligned surfaces, basically we are
    stuck with fixed memory with pitch 1024.
  */
  pSavageSubpicture->pitch = 10;

  /*
     offsets[0] == offset into the map described by either
     address (Client memeory address) or offset (physical offset from fb base)
  */

  pSavageSubpicture->offsets[0] = (unsigned int)malloc(1024*768);

  if(pSavageSubpicture->offsets[0] == (unsigned int)NULL) {
    printf("CreateSubPicture:No system memory\n");
    free(pSavageSubpicture);
    return BadAlloc;
  }


  /* Clear the surface to 0 */
  memset((void *)((unsigned int)pSavageSubpicture->data +
		  (unsigned int)pSavageSubpicture->offsets[0]),0,
	 ((1<<pSavageSubpicture->pitch) * subpicture->height));

  pSavageXvMC->ref++;
#ifdef XVMCLOG
  printf("End XvMCCreateSubpicture\n");
#endif
  return Success;
}



/***************************************************************************
// Function: XvMCClearSubpicture
// Description: Clear the area of the given subpicture to "color".
//		structure passed to it and returning Success.
// Arguments:
//   display - Connection to the X server.
//   subpicture - Subpicture to clear.
//   x, y, width, height - rectangle in the subpicture to clear.
//   color - The data to file the rectangle with.
//
// Returns: Status
***************************************************************************/
Status XvMCClearSubpicture(Display *display, XvMCSubpicture *subpicture,
			  short x, short y,
			  unsigned short width, unsigned short height,
			  unsigned int color) {

  savageXvMCContext *pSavageXvMC;
  savageXvMCSubpicture *pSavageSubpicture;
  int i;
#ifdef XVMCLOG
  printf("Start XvMCClearSubpicture\n");
#endif
  if((subpicture == NULL) || (display == NULL)){
    return BadValue;
  }

  if(!subpicture->privData) {
    return (error_base + XvMCBadSubpicture);
  }
  pSavageSubpicture = (savageXvMCSubpicture *)subpicture->privData;

  pSavageXvMC = (savageXvMCContext *)pSavageSubpicture->privContext;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadSubpicture);
  }

  if((x < 0) || (x + width > subpicture->width)) {
    return BadValue;
  }

  if((y < 0) || (y + height > subpicture->height)) {
    return BadValue;
  }

  for(i=y; i<y + height; i++) {
    memset((void *)((uint)pSavageSubpicture->data +
		    (uint)pSavageSubpicture->offsets[0] + x +
		    (1<<pSavageSubpicture->pitch) * i),(char)color,width);
  }

  pSavageXvMC->dwFlags |= SUBPICIMAGE_CHANGED;
#ifdef XVMCLOG
  printf("End XvMCClearSubpicture\n");
#endif
  return Success;
}

/***************************************************************************
// Function: XvMCCompositeSubpicture
// Description: Composite the XvImae on the subpicture. This composit uses
//		non-premultiplied alpha. Destination alpha is utilized
//		except for with indexed subpictures. Indexed subpictures
//		use a simple "replace".
// Arguments:
//   display - Connection to the X server.
//   subpicture - Subpicture to clear.
//   image - the XvImage to be used as the source of the composite.
//   srcx, srcy, width, height - The rectangle from the image to be used.
//   dstx, dsty - location in the subpicture to composite the source.
//
// Returns: Status
***************************************************************************/
Status XvMCCompositeSubpicture(Display *display, XvMCSubpicture *subpicture,
			       XvImage *image,
			       short srcx, short srcy,
			       unsigned short width, unsigned short height,
			       short dstx, short dsty) {

  savageXvMCContext *pSavageXvMC;
  savageXvMCSubpicture *pSavageSubpicture;
  int i;

#ifdef XVMCLOG
  printf("Start XvMCCompositeSubpicture\n");
#endif
  if((subpicture == NULL) || (display == NULL)){
    printf("Bad subpicture and display\n");
    return BadValue;
  }

  if(!subpicture->privData) {
    printf("Bad privatedata\n");
    return (error_base + XvMCBadSubpicture);
  }
  pSavageSubpicture = (savageXvMCSubpicture *)subpicture->privData;

  pSavageXvMC = (savageXvMCContext *)pSavageSubpicture->privContext;
  if(pSavageXvMC == NULL) {
    printf("Bad context\n");
    return (error_base + XvMCBadSubpicture);
  }

  if((srcx < 0) || (srcx + width > image->width)) {
    printf("Bad srcx=%d, srcwidth=%d, imagewidth=%d\n", srcx, width,image->width);
    return BadValue;
  }

  if((dstx < 0) || (dstx + width > subpicture->width)) {
    printf("Bad dstx\n");
    return BadValue;
  }

  if((srcy < 0) || (srcy + height > image->height)) {
    printf("Bad srcy\n");
    return BadValue;
  }

  if((dsty < 0) || (dsty + height > subpicture->height)) {
    printf("Bad dsty\n");
    return BadValue;
  }

  for(i=0; i<height; i++) {
    memcpy((void *)((uint)pSavageSubpicture->data +
		    (uint)pSavageSubpicture->offsets[0] + dstx +
		    (1<<pSavageSubpicture->pitch) * (i + dsty)),
	   (void *)((uint)image->data +
		    (uint)image->offsets[0] + srcx +
		    image->pitches[0] * (i + srcy))
	   ,width);
  }

  pSavageXvMC->dwFlags |= SUBPICIMAGE_CHANGED;

#ifdef XVMCLOG
  printf("End XvMCCompositeSubpicture\n");
#endif
  return Success;

}


/***************************************************************************
// Function: XvMCDestroySubpicture
// Description: Destroys the specified subpicture.
// Arguments:
//   display - Connection to the X server.
//   subpicture - Subpicture to be destroyed.
//
// Returns: Status
***************************************************************************/
Status XvMCDestroySubpicture(Display *display, XvMCSubpicture *subpicture) {

  savageXvMCSubpicture *pSavageSubpicture;
  savageXvMCContext *pSavageXvMC;

#ifdef	XVMCLOG
  printf("Start XvMCDestroySubpicture\n");
#endif
  if((display == NULL) || (subpicture == NULL)) {
    return BadValue;
  }
  if(!subpicture->privData) {
    return (error_base + XvMCBadSubpicture);
  }
  pSavageSubpicture = (savageXvMCSubpicture *)subpicture->privData;

  pSavageXvMC = (savageXvMCContext *)pSavageSubpicture->privContext;
  if(!pSavageXvMC) {
    return (error_base + XvMCBadSubpicture);
  }


  if(pSavageSubpicture->last_render) {
    XvMCSyncSubpicture(display,subpicture);
  }

  free((void *)pSavageSubpicture->offsets[0]);


  savage_free_privContext(pSavageXvMC);

  free(pSavageSubpicture);
  subpicture->privData = NULL;
#ifdef XVMCLOG
  printf("End XvMCDestroySubpicture\n");
#endif
  return Success;
}


/***************************************************************************
// Function: XvMCSetSubpicturePalette
// Description: Set the subpictures palette
// Arguments:
//   display - Connection to the X server.
//   subpicture - Subpiture to set palette for.
//   palette - A pointer to an array holding the palette data. The array
//     is num_palette_entries * entry_bytes in size.
// Returns: Status
***************************************************************************/

Status XvMCSetSubpicturePalette(Display *display, XvMCSubpicture *subpicture,
				unsigned char *palette) {
 savageXvMCSubpicture *privSubpicture;
 savageXvMCContext *pSavageXvMC;
 int i,j;

#ifdef XVMCLOG
 printf("Start XvMCSetSubpicturePalette\n");
#endif
 if((display == NULL) || (subpicture == NULL)) {
   return BadValue;
 }
 if(subpicture->privData == NULL) {
   return (error_base + XvMCBadSubpicture);
 }
 privSubpicture = (savageXvMCSubpicture *)subpicture->privData;

 pSavageXvMC = (savageXvMCContext *)privSubpicture->privContext;
 j=0;
 for(i=0; i<16; i++) {
   privSubpicture->palette[0][i] = palette[j++];
   privSubpicture->palette[1][i] = palette[j++];
   privSubpicture->palette[2][i] = palette[j++];
 }

 pSavageXvMC->dwFlags |= COLORPALETTE_CHANGED;
#ifdef XVMC
 printf("End XvMCSetSubpicturePalette\n");
#endif

 return Success;
}

/***************************************************************************
// Function: XvMCBlendSubpicture
// Description:
//    The behavior of this function is different depending on whether
//    or not the XVMC_BACKEND_SUBPICTURE flag is set in the XvMCSurfaceInfo.
//    savage only support frontend behavior.
//
//    XVMC_BACKEND_SUBPICTURE not set ("frontend" behavior):
//
//    XvMCBlendSubpicture is a no-op in this case.
//
// Arguments:
//   display - Connection to the X server.
//   subpicture - The subpicture to be blended into the video.
//   target_surface - The surface to be displayed with the blended subpic.
//   source_surface - Source surface prior to blending.
//   subx, suby, subw, subh - The rectangle from the subpicture to use.
//   surfx, surfy, surfw, surfh - The rectangle in the surface to blend
//	blend the subpicture rectangle into. Scaling can ocure if
//	XVMC_SUBPICTURE_INDEPENDENT_SCALING is set.
//
// Returns: Status
***************************************************************************/
Status XvMCBlendSubpicture(Display *display, XvMCSurface *target_surface,
			 XvMCSubpicture *subpicture,
			 short subx, short suby,
			 unsigned short subw, unsigned short subh,
			 short surfx, short surfy,
			 unsigned short surfw, unsigned short surfh) {

  savageXvMCSurface *privTarget;
  savageXvMCContext *pSavageXvMC;
  savageXvMCSubpicture *privSubpicture;

#ifdef XVMCLOG
 printf("Start XvMCBlendSubpicture\n");
#endif
  if(display == NULL) {
    printf("Bad display\n");
    return BadValue;
  }

  if(subpicture == NULL) {
    printf("Bad subpicture\n");
    return (error_base + XvMCBadSubpicture);
  }

  if(target_surface == NULL) {
    printf("Bad target surface\n");
    return (error_base + XvMCBadSurface);
  }

  if((subpicture->xvimage_id != FOURCC_AI44) &&
     (subpicture->xvimage_id != FOURCC_IA44)) {
      printf("Not right id\n");
      return (error_base + XvMCBadSubpicture);
  }

  if(!subpicture->privData) {
    printf("Bad private\n");
    return (error_base + XvMCBadSubpicture);
  }

  privSubpicture = (savageXvMCSubpicture *)subpicture->privData;

  pSavageXvMC = (savageXvMCContext *)privSubpicture->privContext;

  if(pSavageXvMC == NULL) {
    printf("Bad context\n");
    return (error_base + XvMCBadSubpicture);
  }

  if(!target_surface->privData) {
    printf("Bad target privatedata\n");
    return (error_base + XvMCBadSurface);
  }

  privTarget = (savageXvMCSurface *)target_surface->privData;

  /* Check that size isn't bigger than subpicture */
  if((subx + subw) > subpicture->width) {
    printf("Bad subx\n");
    return BadValue;
  }

  if((suby + subh) > subpicture->height) {
    printf("Bad suby\n");
    return BadValue;
  }

  /* Check that dest isn't bigger than surface */
  if((surfx + surfw) > target_surface->width) {
    printf("Bad surfx\n");
    return BadValue;
  }

  if((surfy + surfh) > target_surface->height) {
    printf("Bad surfy\n");
    return BadValue;
  }

#ifdef XVMCLOG
    printf("XvMC: blend. subx: %d, suby: %d, subw: %d, subh: %d, srcx: %d, srcy: %d.\n",
	   subx,suby,subw,subh,srcx,srcy);
#endif

  if (
      (privSubpicture->subx != subx) ||
      (privSubpicture->suby != suby) ||
      (privSubpicture->subw != subw) ||
      (privSubpicture->subh != subh) ||
      (privSubpicture->dstx != surfx) ||
      (privSubpicture->dsty != surfy)
      ) {
    privSubpicture->subx = subx;
    privSubpicture->suby = suby;
    privSubpicture->subw = subw;
    privSubpicture->subh = subh;
    privSubpicture->dstx = surfx;
    privSubpicture->dsty = surfy;

    pSavageXvMC->dwFlags |= SUBPIC_ENABLE | SUBPICAREA_CHANGED;

  } else

    pSavageXvMC->dwFlags |= SUBPIC_ENABLE;

  pSavageXvMC->lpSubpicture = privSubpicture;

#ifdef XVMCLOG
 printf("End XvMCBlendSubpicture\n");
#endif
  return Success;
}



/***************************************************************************
// Function: XvMCBlendSubpicture2
// Description:
//    The behavior of this function is different depending on whether
//    or not the XVMC_BACKEND_SUBPICTURE flag is set in the XvMCSurfaceInfo.
//    savage only supports backend blending.
//
//    XVMC_BACKEND_SUBPICTURE set ("backend" behavior):
//
//    XvMCBlendSubpicture2 blends the source_surface and subpicture and
//    puts it in the target_surface.  This does not effect the status of
//    the source surface but will cause the target_surface to query
//    XVMC_RENDERING until the blend is completed.
//
// Arguments:
//   display - Connection to the X server.
//   subpicture - The subpicture to be blended into the video.
//   target_surface - The surface to be displayed with the blended subpic.
//   source_surface - Source surface prior to blending.
//   subx, suby, subw, subh - The rectangle from the subpicture to use.
//   surfx, surfy, surfw, surfh - The rectangle in the surface to blend
//	blend the subpicture rectangle into. Scaling can ocure if
//	XVMC_SUBPICTURE_INDEPENDENT_SCALING is set.
//
// Returns: Status
***************************************************************************/

/* This func is not fixed as the blend() func above. */

Status XvMCBlendSubpicture2(Display *display,
			  XvMCSurface *source_surface,
			  XvMCSurface *target_surface,
			  XvMCSubpicture *subpicture,
			  short subx, short suby,
			  unsigned short subw, unsigned short subh,
			  short surfx, short surfy,
			  unsigned short surfw, unsigned short surfh) {
  savageXvMCContext *pSavageXvMC;
  savageXvMCSubpicture *privSubpicture;
  savageXvMCSurface *privTarget;
  savageXvMCSurface *privSource;
  savageXvMCSurface privTmp;

#ifdef XVMCLOG
  printf("Start XvMCBlendSubpicture2\n");
#endif
  if(display == NULL) {
    return BadValue;
  }

  if(subpicture == NULL) {
    return (error_base + XvMCBadSubpicture);
  }

  if((target_surface == NULL) || (source_surface == NULL)) {
    return (error_base + XvMCBadSurface);
  }

  if((subpicture->xvimage_id != FOURCC_AI44) &&
     (subpicture->xvimage_id != FOURCC_IA44)) {
      return (error_base + XvMCBadSubpicture);
  }

  if(!subpicture->privData) {
    return (error_base + XvMCBadSubpicture);
  }
  privSubpicture = (savageXvMCSubpicture *)subpicture->privData;

  pSavageXvMC = (savageXvMCContext *)privSubpicture->privContext;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadSubpicture);
  }

  if(!target_surface->privData) {
    return (error_base + XvMCBadSurface);
  }
  privTarget = (savageXvMCSurface *)target_surface->privData;

  if(!source_surface->privData) {
    return (error_base + XvMCBadSurface);
  }

  privSource = (savageXvMCSurface *)source_surface->privData;

  if((subx + subw) > subpicture->width) {
    return BadValue;
  }

  if((suby + subh) > subpicture->height) {
    return BadValue;
  }

  /* Check that dest isn't bigger than surface */
  if((surfx + surfw) > target_surface->width) {
    return BadValue;
  }

  if((surfy + surfh) > target_surface->height) {
    return BadValue;
  }

  privSubpicture->subx = subx;
  privSubpicture->suby = suby;
  privSubpicture->subw = subw;
  privSubpicture->subh = subh;
  privSubpicture->dstx = surfx;
  privSubpicture->dsty = surfy;

  pSavageXvMC->lpSubpicture = privSubpicture;

  memcpy(&privTmp, privTarget, sizeof(savageXvMCSurface));

  memcpy(privTarget, privSource, sizeof(savageXvMCSurface));

  memcpy(privSource, &privTmp, sizeof(savageXvMCSurface));

  target_surface->surface_id = source_surface->surface_id;
  target_surface->context_id = source_surface->context_id;
  target_surface->surface_type_id = source_surface->surface_type_id;
  target_surface->width = source_surface->width;
  target_surface->height = source_surface->height;

  pSavageXvMC->dwFlags |= SUBPIC_ENABLE | SUBPICAREA_CHANGED;

#ifdef XVMCLOG
  printf("End XvMCBlendSubpicture2\n");
#endif
  return Success;
}



/***************************************************************************
// Function: XvMCSyncSubpicture
// Description: This function blocks until all composite/clear requests on
//		the subpicture have been complete.
// Arguments:
//   display - Connection to the X server.
//   subpicture - The subpicture to synchronize
//
// Returns: Status
***************************************************************************/
Status XvMCSyncSubpicture(Display *display, XvMCSubpicture *subpicture) {
  Status ret;
  int stat=0;
#ifdef XVMCLOG
  printf("Start XvMCSyncSubpicture\n");
#endif
  do {
    ret = XvMCGetSubpictureStatus(display,subpicture,&stat);
  }while(!ret && (stat & XVMC_RENDERING));
#ifdef XVMCLOG
  printf("End XvMCSyncSubpicture\n");
#endif
  return ret;
}



/***************************************************************************
// Function: XvMCFlushSubpicture
// Description: This function commits pending composite/clear requests to
//		ensure that they will be completed in a finite amount of
//		time.
// Arguments:
//   display - Connection to the X server.
//   subpicture - The subpicture whos compsiting should be flushed
//
// Returns: Status
// NOTES: savage always dispatches commands so flush is a no-op
***************************************************************************/
Status XvMCFlushSubpicture(Display *display, XvMCSubpicture *subpicture) {
#ifdef XVMCLOG
  printf("Start XvMCFlushSubpicture\n");
#endif
  if(display == NULL) {
    return BadValue;
  }
  if(subpicture == NULL) {
    return (error_base + XvMCBadSubpicture);
  }
#ifdef XVMCLOG
  printf("End XvMCFlushSubpicture\n");
#endif
  return Success;
}



/***************************************************************************
// Function: XvMCGetSubpictureStatus
// Description: This function gets the current status of a subpicture
//
// Arguments:
//   display - Connection to the X server.
//   subpicture - The subpicture whos status is being queried
//   stat - The status of the subpicture. It can be any of the following
//	    OR'd together:
//	    XVMC_RENDERING  - Last composite or clear request not completed
//	    XVMC_DISPLAYING - Suppicture currently being displayed.
//
// Returns: Status
// Notes: savage always blends into a third surface so the subpicture is
//  never actually displaying, only a copy of it is displaying. We only
//  have to worry about the rendering case.
***************************************************************************/
Status XvMCGetSubpictureStatus(Display *display, XvMCSubpicture *subpicture,
			     int *stat) {

  savageXvMCSubpicture *privSubpicture;
  savageXvMCContext *pSavageXvMC;
#ifdef XVMCLOG
  printf("Start XvMCGetSubpictureStatus\n");
#endif
  if((display == NULL) || (stat == NULL)) {
    return BadValue;
  }
  if((subpicture == NULL) || (subpicture->privData == NULL)) {
    return (error_base + XvMCBadSubpicture);
  }
  *stat = 0;
  privSubpicture = (savageXvMCSubpicture *)subpicture->privData;

  pSavageXvMC = (savageXvMCContext *)privSubpicture->privContext;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadSubpicture);
  }

/*  SAVAGE_LOCK(pSavageXvMC,0);

  if(privSubpicture->last_render &&
      (privSubpicture->last_render > GET_RSTATUS(pSavageXvMC))) {
    *stat |= XVMC_RENDERING;
  }
  SAVAGE_UNLOCK(pSavageXvMC);*/
#ifdef XVMCLOG
  printf("End XvMCGetSubpictureStatus\n");
#endif
  return Success;
}

#define NUM_XVMC_ATTRIBUTES 5
static XvAttribute Savage_XVMC_ATTRIBUTES[] = {
  {XvGettable | XvSettable, 0, (1 << 14) -1, "XV_COLORKEY"},
  {XvGettable | XvSettable, -127, +127,	 "XV_BRIGHTNESS"},
  {XvGettable | XvSettable, 0, 255,	 "XV_CONTRAST"},
  {XvGettable | XvSettable, 0, 255,	 "XV_SATURATION"},
  {XvGettable | XvSettable, -180, 180,	 "XV_HUE"}
};

/* Attributes defination:
 *   Don't know if there are standard specification for these attributes,
 *     standard defination is prefered.
 * Following is current defination:
 * ColorKey: 16bit, RGB565 format color code.
 */

/***************************************************************************
// Function: XvMCQueryAttributes
// Description: An array of XvAttributes of size "number" is returned by
//   this function. If there are no attributes, NULL is returned and number
//   is set to 0. The array may be freed with xfree().
//
// Arguments:
//   display - Connection to the X server.
//   context - The context whos attributes we are querying.
//   number - The number of returned atoms.
//
// Returns:
//  An array of XvAttributes.
// Notes:
//   For savage we support these Attributes:
//    XV_COLORKEY: The colorkey value, initialized from the Xv value at
//		   context creation time.
//    XV_BRIGHTNESS
//    XV_CONTRAST
//    XV_SATURATION
***************************************************************************/
XvAttribute *XvMCQueryAttributes(Display *display, XvMCContext *context,
				 int *number) {
  savageXvMCContext *pSavageXvMC;
  XvAttribute *attributes;
#ifdef XVMCLOG
  printf("Start XvMCQueryAttribute\n");
#endif
  if(number == NULL) {
    return NULL;
  }
  if(display == NULL) {
    *number = 0;
    return NULL;
  }
  if(context == NULL) {
    *number = 0;
    return NULL;
  }
  pSavageXvMC = context->privData;
  if(pSavageXvMC == NULL) {
    *number = 0;
    return NULL;
  }

  attributes = (XvAttribute *)malloc(NUM_XVMC_ATTRIBUTES *
				     sizeof(XvAttribute));
  if(attributes == NULL) {
    *number = 0;
    return NULL;
  }

  memcpy(attributes,Savage_XVMC_ATTRIBUTES,(NUM_XVMC_ATTRIBUTES *
					  sizeof(XvAttribute)));

  *number = NUM_XVMC_ATTRIBUTES;
#ifdef XVMCLOG
  printf("End XvMCQueryAttribute\n");
#endif
  return attributes;
}

/***************************************************************************
// Function: XvMCSetAttribute
// Description: This function sets a context-specific attribute.
//
// Arguments:
//   display - Connection to the X server.
//   context - The context whos attributes we are querying.
//   attribute - The X atom of the attribute to be changed.
//   value - The new value for the attribute.
//
// Returns:
//  Status
// Notes:
//   For savage we support these Attributes:
//    XV_COLORKEY: The colorkey value, initialized from the Xv value at
//		   context creation time.
//    XV_BRIGHTNESS
//    XV_CONTRAST
//    XV_SATURATION
***************************************************************************/
Status XvMCSetAttribute(Display *display, XvMCContext *context,
			Atom attribute, int value) {
  savageXvMCContext *pSavageXvMC;
#ifdef XVMCLOG
  printf("Start XvMCSetAttribute\n");
#endif
  if(display == NULL) {
    return BadValue;
  }
  if(context == NULL) {
    return (error_base + XvMCBadContext);
  }
  pSavageXvMC = context->privData;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadContext);
  }
  if(attribute == pSavageXvMC->xv_colorkey) {
    if((value < Savage_XVMC_ATTRIBUTES[0].min_value) ||
       (value > Savage_XVMC_ATTRIBUTES[0].max_value)) {
      return BadValue;
    }
    pSavageXvMC->colorkey = value;
    printf("ColorKey=%x\n", value);
    XvSetPortAttribute(display, context->port, pSavageXvMC->xv_colorkey, value);
    pSavageXvMC->dwFlags |= COLORKEYCHANGED;
#ifdef XVMCLOG
    printf("End XvMCSetAttribute\n");
#endif
    return Success;
  }
  if(attribute == pSavageXvMC->xv_brightness) {
    if((value < Savage_XVMC_ATTRIBUTES[1].min_value) ||
       (value > Savage_XVMC_ATTRIBUTES[1].max_value)) {
      return BadValue;
    }
    pSavageXvMC->brightness = value;
    XvSetPortAttribute(display, context->port, pSavageXvMC->xv_brightness, value);
    printf("End XvMCSetAttribute\n");
    return Success;
  }
  if(attribute == pSavageXvMC->xv_saturation) {
    if((value < Savage_XVMC_ATTRIBUTES[2].min_value) ||
       (value > Savage_XVMC_ATTRIBUTES[2].max_value)) {
      return BadValue;
    }
    pSavageXvMC->saturation = value;
    XvSetPortAttribute(display, context->port, pSavageXvMC->xv_saturation, value);
    printf("End XvMCSetAttribute\n");
    return Success;
  }
  if(attribute == pSavageXvMC->xv_contrast) {
    if((value < Savage_XVMC_ATTRIBUTES[3].min_value) ||
       (value > Savage_XVMC_ATTRIBUTES[3].max_value)) {
      return BadValue;
    }
    pSavageXvMC->contrast = value;
    XvSetPortAttribute(display, context->port, pSavageXvMC->xv_contrast, value);
#ifdef XVMCLOG
    printf("End XvMCSetAttribute\n");
#endif
    return Success;
  }

  if(attribute == pSavageXvMC->xv_hue) {
    if((value < Savage_XVMC_ATTRIBUTES[4].min_value) ||
       (value > Savage_XVMC_ATTRIBUTES[4].max_value)) {
      return BadValue;
    }
    pSavageXvMC->hue = value;
    XvSetPortAttribute(display, context->port, pSavageXvMC->xv_hue, value);
#ifdef XVMCLOG
    printf("End XvMCSetAttribute\n");
#endif
    return Success;
  }
#ifdef XVMCLOG
  printf("End XvMCSetAttribute\n");
#endif
  return BadValue;
}

/***************************************************************************
// Function: XvMCGetAttribute
// Description: This function queries a context-specific attribute and
//   returns the value.
//
// Arguments:
//   display - Connection to the X server.
//   context - The context whos attributes we are querying.
//   attribute - The X atom of the attribute to be queried
//   value - The returned attribute value
//
// Returns:
//  Status
// Notes:
//   For savage we support these Attributes:
//    XV_COLORKEY: The colorkey value, initialized from the Xv value at
//		   context creation time.
//    XV_BRIGHTNESS
//    XV_CONTRAST
//    XV_SATURATION
***************************************************************************/
Status XvMCGetAttribute(Display *display, XvMCContext *context,
			Atom attribute, int *value) {
  savageXvMCContext *pSavageXvMC;
#ifdef XVMCLOG
  printf("Start XvMCGetAttribute\n");
#endif
  if(display == NULL) {
    return BadValue;
  }
  if(context == NULL) {
    return (error_base + XvMCBadContext);
  }
  pSavageXvMC = context->privData;
  if(pSavageXvMC == NULL) {
    return (error_base + XvMCBadContext);
  }
  if(value == NULL) {
    return BadValue;
  }

  if(attribute == pSavageXvMC->xv_colorkey) {
    *value = pSavageXvMC->colorkey;
    return Success;
  }
  if(attribute == pSavageXvMC->xv_brightness) {
    *value = pSavageXvMC->brightness;
    return Success;
  }
  if(attribute == pSavageXvMC->xv_saturation) {
    *value = pSavageXvMC->saturation;
    return Success;
  }
  if(attribute == pSavageXvMC->xv_contrast) {
    *value = pSavageXvMC->contrast;
    return Success;
  }
  if(attribute == pSavageXvMC->xv_hue) {
    *value = pSavageXvMC->hue;
    return Success;
  }
#ifdef XVMCLOG
  printf("End XvMCGetAttribute\n");
#endif
  return BadValue;
}








