/*
 * Copyright 2003 VIA Technologies, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/***************************************************************************
 * libsavageXvMC.h: MC Driver savage includes
 *
 * Authors:
 *	John Zhao <jzhao@s3graphics.com>
 *
 *
 ***************************************************************************/
/* $XFree86: xc/lib/XvMC/hw/savage/savageXvMC.h,v 1.2 2001/12/18 04:23:56 dawes Exp $ */


#ifndef ISAVAGEXVMC_H
#define ISAVAGEXVMC_H

/* #define XVMC_DEBUG(x) do {x; }while(0); */
#define XVMC_DEBUG(x)

#include <xf86drm.h>
#include <Xlibint.h>

#define MAXFLIPBUFF	 2

typedef struct _savageOverlay {
    unsigned int ssControl;
    unsigned int srcx;
    unsigned int srcy;
    unsigned int srcWidth;
    unsigned int srcHeight;
    unsigned int destx;
    unsigned int desty;
    unsigned int destWidth;
    unsigned int destHeight;
    unsigned short wWidth;
    unsigned short wHeight;
    unsigned int sw,dw;	 /* Left_Out Original arguments passed into PutSurface(), for scaling precision. */
}  savageOverlay;

/***************************************************************************
// This structure is referenced by the privData pointer in the XvMCSubpicture
//  structure.
***************************************************************************/
typedef struct _SUBPICINFO  /* Info for texture. Blend will use this when setting hw. */
{
	/* Currently only dwTexDesc, fDestW and fDestH are used. */
    unsigned short     srcWidth;
    unsigned short     srcHeight;
    unsigned int       dwSubWidth;
    unsigned int       dwSubHeight;
    unsigned int       dwDestX;
    float	       fDestX;
    unsigned int       dwDestY;
    float	       fDestY;
    unsigned int       dwDestW;
    float	       fDestW;
    unsigned int       dwDestH;
    float	       fDestH;
    float	       fTexX;
    float	       fTexY;
    unsigned int       dwTexDesc;
} SUBPICINFO;

typedef struct _savageXvMCSubpicture {
  unsigned int pitch;
  unsigned int last_render;
  unsigned int last_flip;
  drmAddress data;
  drmHandle handle;
  unsigned int offsets[1];
  unsigned char palette[3][16];
  unsigned short subx, suby; /* blend src from subpicture */
  unsigned short subw, subh; /* blend src from subpicture */
  unsigned short dstx, dsty; /* blend dst in dst surface */
  void *privContext;
  SUBPICINFO pInfo;
} savageXvMCSubpicture;

/***************************************************************************
// savageXvMCDrmMap: Holds the data about the DRM maps
***************************************************************************/
typedef struct _savageXvMCDrmMap {
  drmHandle handle;
  unsigned int offset;
  drmAddress address;
  unsigned int size;
} savageXvMCDrmMap, *savageXvMCDrmMapPtr;

typedef struct _savageXvMCDCTBlock {
  unsigned int offset;
  unsigned int used;
  unsigned short dwTag;
  unsigned short dwOldTag;
} savageXvMCDCTBlock, *savageXvMCDCTBlockPtr;
/***************************************************************************
//  pointer in the XvMCContext structure.
***************************************************************************/
typedef struct _savageXvMCContext {
  int fd;   /* File descriptor for /dev/dri */
  savageXvMCDrmMap MMIO;
  savageXvMCDrmMap DCTBlocks;
  savageXvMCDrmMap Surfaces;
  savageXvMCDrmMap Aperture;
  drmContext drmcontext;
  unsigned int last_render;
  unsigned int last_flip;
  unsigned short ref;
  unsigned short current;
  int lock;   /* Lightweight lock to avoid locking twice */
  char busIdString[10]; /* PCI:0:1:0 or PCI:0:2:0 */
  unsigned int dual_prime; /* Flag to identify when dual prime is in use. */
  savageXvMCDCTBlock DCTBlock[4];
  unsigned int fb_base;		      /* Frame buffer base physical address */
  unsigned int BCIBaseAddr;
  unsigned int *lpSymbolBuffer;
  unsigned int FlipBuff[MAXFLIPBUFF]; /* Destination flip buffer physical address*/
  unsigned int dwCurBuff;
  unsigned int dwSubPicMixer;	      /* Temparary mixer buffer physical address*/
  unsigned int dwMixerStride;	      /* Temparary mixer buffer stride*/
  unsigned int dwStride;	      /* Flip buffer stride*/
  unsigned int SubPicBuff;	      /* Subpicture buffer linear address*/
  unsigned int SubPicPalette;	      /* Subpicture palette buffer linear address*/
  unsigned int SubPicBuffOffset;      /* Subpicture buffer physical address*/
  unsigned int SubPicPaletteOffset;   /* Subpicture palette buffer physical address*/
  unsigned int dwIDCTBuffUsed;
  unsigned char * lpIDCTBuff;
  unsigned short wWidth;
  unsigned short wHeight;
  unsigned int dwPictureType;
  unsigned int dwFlags;
  unsigned int dwDummyTile;
  savageXvMCSubpicture * lpSubpicture;
  savageOverlay Overlay;
  unsigned int dwTiledSurfReg;
  unsigned int fpTargetAperture;
  unsigned int depth;
  unsigned int displayWidth;
  unsigned int TileOffset;
  unsigned int frameX0;
  unsigned int frameY0;
  unsigned int IOBase;
  unsigned short usEventTag;
  Atom xv_colorkey;
  Atom xv_brightness;
  Atom xv_contrast;
  Atom xv_saturation;
  Atom xv_hue;
  unsigned int	port;
  GC		gc;
  XvImage     *	 image;
  char	      *	 data;
  int brightness;
  int saturation;
  int contrast;
  int hue;
  int colorkey;
  unsigned int dwOverFlow;
  unsigned int * pjShadowMemory;
} savageXvMCContext;

/***************************************************************************
// savageXvMCSurface: Private data structure for each XvMCSurface. This
//  structure is referenced by the privData pointer in the XvMCSurface
//  structure.
***************************************************************************/
typedef struct _savageXvMCSurface {
  unsigned int pitch;
  unsigned int last_render;
  unsigned int last_flip;
  unsigned int second_field; /* Flags a surface that is only half done */
  drmAddress data;
  drmHandle handle;
  unsigned int offsets[1];
  unsigned short srcx, srcy;
  unsigned short srcw, srch;
  unsigned short destx, desty;
  unsigned short destw, desth;
  unsigned int dwDisplayTaskType;
  unsigned short wWidth;
  unsigned short wHeight;
  unsigned int dwLinearPitch;
  unsigned int dwDummyLine;
  unsigned int pType;
  savageXvMCContext *privContext;
} savageXvMCSurface;


typedef struct _Box {
  int x1,y1,x2,y2;
} Box, *BoxPtr;

typedef struct	_MCS3BeginData {
  unsigned int dwFrameDesc;
  unsigned short width;
  unsigned short height;
} MCS3BeginData, *LPMCS3BeginData;
/***************************************************************************
// drm_savage_overlay_t: Structure returned by overlay info ioctl.
//  structure in the kernel.
***************************************************************************/
typedef struct _drm_savage_overlay_t {
  unsigned int offset;
  unsigned int physical;
} drm_savage_overlay_t;

/* Subpicture fourcc */
#define FOURCC_IA44 0x34344149

/* Static Parameters */
#define I810_XVMC_MAXWIDTH  720
#define I810_XVMC_MAXHEIGHT  576
#define I810_DEFAULT16_COLORKEY 31

#define RGB16ToColorKey(c) \
	(((c & 0xF800) << 8) | ((c & 0x07E0) << 5) | ((c & 0x001F) << 3))

/* Locking Macros lightweight lock used to prevent relocking */
#define SAVAGE_LOCK(c,f)		    \
    drmGetLock(c->fd, c->drmcontext, f);

#define SAVAGE_UNLOCK(c)		     \
    drmUnlock(c->fd, c->drmcontext);

#if 0
#define SAVAGE_LOCK(c,f)		     \
  if(!c->lock) {			   \
    drmGetLock(c->fd, c->drmcontext, f);   \
  }					   \
  c->lock++;

#define SAVAGE_UNLOCK(c)		     \
  c->lock--;				   \
  if(!c->lock) {			   \
    drmUnlock(c->fd, c->drmcontext);	   \
  }
#endif

#define FTYPE_I	       1
#define FTYPE_P	       2
#define FTYPE_B	       3

#define FRAME_I_PIC		       (0x1<<27) /* I picture will be decoded */
#define FRAME_P_PIC		       (0x2<<27) /* P picture will be decoded */
#define FRAME_B_PIC		       (0x3<<27) /* B picture will be decoded */
#define FRAME_TOP_FIELD		       (0x1<<29) /* Top field will be decoded */
						 /* only for field picture */
#define FRAME_BOTTOM_FIELD	       (0x2<<29) /* Bottom field will be */
						 /* decoded only for field */
						 /* picture */
#define FRAME_FRAME		       (0x3<<29) /* Frame will be decoded */
						 /* only for frame picture */

#define FRAME_FIRST_P_FIELD	       0x0
#define FRAME_SECOND_P_FIELD	       (0x1<<31)

#define MB_DECODE_CrBLK			0x1	    /* Cr block is coded */
#define MB_DECODE_CbBLK			(0x1<<1)    /* Cb block is coded */
#define MB_DECODE_Y3BLK			(0x1<<2)    /* Y3 block is coded */
#define MB_DECODE_Y2BLK			(0x1<<3)    /* Y2 block is coded */
#define MB_DECODE_Y1BLK			(0x1<<4)    /* Y1 block is coded */
#define MB_DECODE_Y0BLK			(0x1<<5)    /* Y0 block is coded */
#define MB_DCT_FRAME			0x0	    /* Frame DCT	 */
#define MB_DCT_FIELD			(0x1<<6)    /* Filed DCT	 */
#define MB_TYPE_INTRA			0x0	    /* Intra MacroBlock */
#define MB_TYPE_INTER_F			(0x1<<7)    /* Inter MacroBlock */
						    /* Forward Prediction */
#define MB_TYPE_INTER_B			(0x2<<7)    /* Inter MacroBlock */
						    /* Backward Prediction */
#define MB_TYPE_INTER_FB		(0x3<<7)    /* Inter MacroBlock */
						    /* Forward and Backword */
						    /* Prediction */
#define MB_MOTION_FIELD			(0x1<<9)    /* Filed Prediction */
#define MB_MOTION_FRAME			(0x2<<9)    /* Frame or 16x8 Prediction */
#define MB_MOTION_DP			(0x3<<9)    /* Dual prime Prediction */
#define MB_MV3_DISABLED			0x0	    /* Disable Motion Vector 3 */
#define MB_MV3_ENABLED			(0x1<<23)   /* Enable Motion Vector 3  */
#define MB_MV2_DISABLED			0x0	    /* Disable Motion Vector 2 */
#define MB_MV2_ENABLED			(0x1<<24)   /* Enable Motion Vector 2  */
#define MB_MV1_DISABLED			0x0	    /* Disable Motion Vector 1 */
#define MB_MV1_ENABLED			(0x1<<25)   /* Enable Motion Vector 1  */
#define MB_MV0_DISABLED			0x0	    /* Disable Motion Vector 0 */
#define MB_MV0_ENABLED			(0x1<<26)   /* Enable Motion Vector 0  */
#define MB_CR_OVERFLOWED		(0x1<<16)   /* Cr block overflowed */
#define MB_CB_OVERFLOWED		(0x1<<17)   /* Cb block overflowed */
#define MB_Y3_OVERFLOWED		(0x1<<18)   /* Y3 block overflowed */
#define MB_Y2_OVERFLOWED		(0x1<<19)   /* Y2 block overflowed */
#define MB_Y1_OVERFLOWED		(0x1<<20)   /* Y1 block overflowed */
#define MB_Y0_OVERFLOWED		(0x1<<21)   /* Y0 block overflowed */
#define MB_COMBINED_IDCT		(0x1<<27)   /* Indicates that any overflowed */

#define MB_SUBTRACT_OFIDCT		(0x1<<28)   /* Indicates that Overflow IDCT*/
						       /* will be subtracted rather than*/
						       /* added*/
#define MB_OVERFLOW_PACKED		(0x1<<29)   /* Indicates that Overflow*/
						       /* data are packed right after*/
							  /* saturated data , rather than*/
							     /* at fixed offset 384 (6 blocks*/
								/* after saturated data)*/
#define MB_IDCT_16BPP			(0x1<<30)   /* Signed*/
#define MB_IDCT_SIGNED			(0x1<<31)   /* Not needed when MB_IDCT_16BPP is set*/
						       /* after saturated data)*/
#define MV_BOTTOM_REF_FIELD		 (0X1<<12)

#define MB_9B_ENABLED			 (0x1L<<27)

#define DISPLAY_FRAME			 (0x1 << 0)
#define DISPLAY_TOP_FIELD		 (0x1 << 1)
#define DISPLAY_BOTTOM_FIELD		 (0x1 << 2)

#define	   SUBPIC_ENABLE	    0x1 /*Enable sub picture for the immediately*/
					/*display frame*/
#define	   COLORPALETTE_CHANGED	    0x2 /*Color palette is changed compare with*/
					/*the previous frame*/
#define	   SUBPICIMAGE_CHANGED	    0x4 /*The bitmap of sub picture is changed*/
					/*compare with the previous frame*/
#define	   SUBPICAREA_CHANGED	    0x8 /*The sub picture starting position*/
					/*or width or height is changed*/
#define	   MB_DUMMY_BLOCK	    0x10
#define	   STREAM_ON		    0x20

#define	   PALETTECHANGED	    0x80000000
#define	   COLORKEYCHANGED	    0x40000000

typedef struct _MCS3MacroBLKData
{
    unsigned int dwMacroBLKDesc;/* IDCT, DCT type,  Prediction direction,*/
				/* Predition type, Motion vectors information */
				/* for current macro block */
    unsigned int dwMBLKHOff;	/* 0x0 is the first column macroblocks, 0x10 is*/
				/* the second */
    unsigned int dwMBLKVOff;	/* 0x0 is the fisrt row macroblocks, 0x10 is */
				/* the second */
    unsigned short  wMV[8];	/* 4 motion vectors */
				/* w[0] is vector[0][0][0] */
				/* w[1] is vector[0][0][1] */
				/* w[2] is vector[1][0][0] */
				/* w[3] is vector[1][0][1] */
				/* w[4] is vector[0][1][0] */
				/* w[5] is vector[0][1][1] */
				/* w[6] is vector[1][1][0] */
				/* w[7] is vector[1][1][1] */
    void * lpIDCTBuff;		/* IDCT data for current macro block */
				/* Client need to alloc and maintain it */
} MCS3MacroBLKData, *LPMCS3MacroBLKData;

/*
  Definitions for temporary wire protocol hooks to be replaced
  when a HW independent libXvMC is created.
*/
extern Status _xvmc_create_context(Display *dpy, XvMCContext *context,
				   int *priv_count, uint **priv_data);

extern Status _xvmc_destroy_context(Display *dpy, XvMCContext *context);

extern Status _xvmc_create_surface(Display *dpy, XvMCContext *context,
				   XvMCSurface *surface, int *priv_count,
				   uint **priv_data);

extern Status _xvmc_destroy_surface(Display *dpy, XvMCSurface *surface);

extern Status  _xvmc_create_subpicture(Display *dpy, XvMCContext *context,
				       XvMCSubpicture *subpicture,
				       int *priv_count, uint **priv_data);

extern Status	_xvmc_destroy_subpicture(Display *dpy,
					 XvMCSubpicture *subpicture);

/*
  Prototypes
*/
#endif







































































































